from autolens.pixelization import pixelization

import pytest
import numpy as np

from autolens.imaging import mask
from test.mock.mock_mask import MockSubGridCoords, MockGridCollection, MockBorderCollection


def grid_to_pix_pixels_via_nearest_neighbour(grid, pix_centers):
    def compute_squared_separation(coordinate1, coordinate2):
        """Computes the squared separation of two image_grid (no square root for efficiency)"""
        return (coordinate1[0] - coordinate2[0]) ** 2 + (coordinate1[1] - coordinate2[1]) ** 2

    image_pixels = grid.shape[0]

    image_to_pix = np.zeros((image_pixels,))

    for image_index, image_coordinate in enumerate(grid):
        distances = list(map(lambda centers: compute_squared_separation(image_coordinate, centers),
                             pix_centers))

        image_to_pix[image_index] = (np.argmin(distances))

    return image_to_pix


class MockSparseMask(object):

    def __init__(self, sparse_to_image, image_to_sparse):
        """ The KMeans clustering used to derive an amorphous pixelization uses a set of image-grid grid. For \
        high resolution imaging, the large number of grid makes KMeans clustering (unfeasibly) slow.

        Therefore, for efficiency, we define a 'clustering-grid', which is a sparsely sampled set of image-grid \
        grid used by the KMeans algorithm instead. However, we don't need the actual grid of this \
        clustering grid (as they are already calculated for the image-grid). Instead, we just need a mapper between \
        clustering-data_to_image and image-data_to_image.

        Thus, the *cluster_to_image* attribute maps every pixel on the clustering grid to its closest image pixel \
        (via the image pixel's 1D index). This is used before the KMeans clustering algorithm, to extract the sub-set \
        of grid that the algorithm uses.

        By giving the KMeans algorithm only clustering-grid grid, it will only tell us the mappings between \
        pix-data_to_image and clustering-data_to_image. However, to perform the pix reconstruction, we need to
        know all of the mappings between pix data_to_image and image data_to_image / sub-image data_to_image. This
        would require a (computationally expensive) nearest-neighbor search (over all clustering data_to_image and
        image / sub data_to_image) to calculate. The calculation can be sped-up by using the attribute
        *image_to_cluster*, which maps every image-pixel to its closest pixel on the clustering grid (see
        *pixelization.sub_grid_to_pix_pixels_via_sparse_pairs*).
        """

        self.sparse_to_image = sparse_to_image
        self.image_to_sparse = image_to_sparse


class MockGeometry(object):

    def __init__(self, y_min, y_max, x_min, x_max, y_pixel_scale, x_pixel_scale):
        self.y_min = y_min
        self.y_max = y_max
        self.x_min = x_min
        self.x_max = x_max
        self.y_pixel_scale = y_pixel_scale
        self.x_pixel_scale = x_pixel_scale


@pytest.fixture(name="five_pixels")
def make_five_pixels():
    return np.array([[0, 0], [0, 1], [1, 0], [1, 1], [1, 2]])


@pytest.fixture(name="three_pixels")
def make_three_pixels():
    return np.array([[0, 0], [0, 1], [1, 0]])


class TestPixelizationGrid:

    class TestCoordinateGridWithinAnnulus:

        def test__shape_3x3__circle_radius_15__all_9_pixels_in_grid_with_correct_coordinates(self):

            pix_grid = pixelization.PixelizationGrid(shape=(3,3))

            coordinate_grid = pix_grid.coordinate_grid_within_annulus(inner_radius=0.0, outer_radius=1.5)

            assert (coordinate_grid == np.array([[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0],
                                                  [0.0, -1.0],  [0.0, 0.0],  [0.0, 1.0],
                                                  [1.0, -1.0],  [1.0, 0.0],  [1.0, 1.0]])).all()

        def test__shape_3x3__circle_radius_3__all_9_pixels_in_grid_with_correct_coordinates(self):

            pix_grid = pixelization.PixelizationGrid(shape=(3,3))

            coordinate_grid = pix_grid.coordinate_grid_within_annulus(inner_radius=0.0, outer_radius=3)

            assert (coordinate_grid == np.array([[-2.0, -2.0], [-2.0, 0.0], [-2.0, 2.0],
                                                  [0.0, -2.0],  [0.0, 0.0],  [0.0, 2.0],
                                                  [2.0, -2.0],  [2.0, 0.0],  [2.0, 2.0]])).all()

        def test__shape_3x2__circle_radius_15__all_6_pixels_in_grid_with_correct_coordinates(self):

            pix_grid = pixelization.PixelizationGrid(shape=(3, 2))

            coordinate_grid = pix_grid.coordinate_grid_within_annulus(inner_radius=0.0, outer_radius=1.5)

            assert (coordinate_grid == np.array([[-1.0, -0.75], [-1.0, 0.75],
                                                  [0.0, -0.75],  [0.0, 0.75],
                                                  [1.0, -0.75],  [1.0, 0.75]])).all()

        def test__shape_2x3__circle_radius_15__all_6_pixels_in_grid_with_correct_coordinates(self):

            pix_grid = pixelization.PixelizationGrid(shape=(2, 3))

            coordinate_grid = pix_grid.coordinate_grid_within_annulus(inner_radius=0.0, outer_radius=1.5)

            assert (coordinate_grid == np.array([[-0.75, -1.0], [-0.75, 0.0], [-0.75, 1.0],
                                                  [0.75, -1.0],  [0.75, 0.0], [0.75, 1.0]])).all()



class TestPixelization:


    class TestMappingMatrix:

        def test__3_image_pixels__6_pix_pixels__sub_grid_1x1(self, three_pixels):
            sub_to_pix = np.array([0, 1, 2])
            sub_to_image = np.array([0, 1, 2])

            grids = MockGridCollection(image=three_pixels,
                                       sub=MockSubGridCoords(three_pixels,
                                                                   sub_to_image,
                                                                   sub_grid_size=1))

            pix = pixelization.Pixelization(pixels=6)
            mapping_matrix = pix.mapping_matrix_from_sub_to_pix_jit(sub_to_pix, grids)

            assert (mapping_matrix == np.array([[1, 0, 0, 0, 0, 0],  # Image pixel 0 maps to pix pixel 0.
                                                [0, 1, 0, 0, 0, 0],  # Image pixel 1 maps to pix pixel 1.
                                                [0, 0, 1, 0, 0, 0]])).all()  # Image pixel 2 maps to pix pixel 2

        def test__5_image_pixels__8_pix_pixels__sub_grid_1x1(self, five_pixels):
            sub_to_pix = np.array([0, 1, 2, 7, 6])
            sub_to_image = np.array([0, 1, 2, 3, 4])

            grids = MockGridCollection(
                image=five_pixels,
                sub=MockSubGridCoords(five_pixels, sub_to_image,
                                      sub_grid_size=1))

            pix = pixelization.Pixelization(pixels=8)
            mapping_matrix = pix.mapping_matrix_from_sub_to_pix_jit(sub_to_pix, grids)

            assert (mapping_matrix == np.array(
                [[1, 0, 0, 0, 0, 0, 0, 0],  # Image image_to_pixel 0 and 3 map to pix pixel 0.
                 [0, 1, 0, 0, 0, 0, 0, 0],  # Image image_to_pixel 1 and 4 map to pix pixel 1.
                 [0, 0, 1, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0, 0, 1],
                 [0, 0, 0, 0, 0, 0, 1, 0]])).all()  # Image image_to_pixel 2 and 5 map to pix pixel 2

        def test__5_image_pixels__8_pix_pixels__sub_grid_2x2__no_overlapping_pixels(self, five_pixels):
            sub_to_pix = np.array([0, 1, 2, 3, 1, 2, 3, 4, 2, 3, 4, 5, 7, 0, 1, 3, 6, 7, 4, 2])
            sub_to_image = np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4])

            grids = MockGridCollection(image=five_pixels,
                                       sub=MockSubGridCoords(five_pixels, sub_to_image,
                                                                   sub_grid_size=2))

            pix = pixelization.Pixelization(pixels=8)
            mapping_matrix = pix.mapping_matrix_from_sub_to_pix_jit(sub_to_pix, grids)

            assert (mapping_matrix == np.array(
                [[0.25, 0.25, 0.25, 0.25, 0, 0, 0, 0],
                 [0, 0.25, 0.25, 0.25, 0.25, 0, 0, 0],
                 [0, 0, 0.25, 0.25, 0.25, 0.25, 0, 0],
                 [0.25, 0.25, 0, 0.25, 0, 0, 0, 0.25],
                 [0, 0, 0.25, 0, 0.25, 0, 0.25, 0.25]])).all()

        def test__5_image_pixels__8_pix_pixels__sub_grid_2x2__include_overlapping_pixels(self, five_pixels):
            sub_to_pix = np.array([0, 0, 0, 1, 1, 1, 0, 0, 2, 3, 4, 5, 7, 0, 1, 3, 6, 7, 4, 2])
            sub_to_image = np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4])

            grids = MockGridCollection(image=five_pixels,
                                       sub=MockSubGridCoords(five_pixels, sub_to_image,
                                                                   sub_grid_size=2))

            pix = pixelization.Pixelization(pixels=8)
            mapping_matrix = pix.mapping_matrix_from_sub_to_pix_jit(sub_to_pix, grids)

            assert (mapping_matrix == np.array(
                [[0.75, 0.25, 0, 0, 0, 0, 0, 0],
                 [0.5, 0.5, 0, 0, 0, 0, 0, 0],
                 [0, 0, 0.25, 0.25, 0.25, 0.25, 0, 0],
                 [0.25, 0.25, 0, 0.25, 0, 0, 0, 0.25],
                 [0, 0, 0.25, 0, 0.25, 0, 0.25, 0.25]])).all()

        def test__3_image_pixels__6_pix_pixels__sub_grid_4x4(self, three_pixels):
            sub_to_pix = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1,
                                   2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
                                   0, 1, 2, 3, 4, 5, 0, 1, 2, 3, 4, 5, 0, 1, 2, 3])

            sub_to_image = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                                     2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2])

            grids = MockGridCollection(image=three_pixels,
                                       sub=MockSubGridCoords(three_pixels,
                                                                   sub_to_image, sub_grid_size=4))

            pix = pixelization.Pixelization(pixels=6)
            mapping_matrix = pix.mapping_matrix_from_sub_to_pix_jit(sub_to_pix, grids)

            assert (mapping_matrix == np.array(
                [[0.75, 0.25, 0, 0, 0, 0],
                 [0, 0, 1.0, 0, 0, 0],
                 [0.1875, 0.1875, 0.1875, 0.1875, 0.125, 0.125]])).all()


class TestRectangularPixelization:

    class TestConstructor:

        def test__number_of_pixels_and_regularization_set_up_correctly(self):
            pix = pixelization.Rectangular(shape=(3, 3), regularization_coefficients=(2.0,))

            assert pix.shape == (3, 3)
            assert pix.pixels == 9
            assert pix.regularization_coefficients == (2.0,)

    class TestSetupGeometry:

        def test__3x3_grid__buffer_is_small__grid_give_min_minus_1_max_1__sets_up_geometry_correctly(self):

            pix = pixelization.Rectangular(shape=(3, 3), regularization_coefficients=(1.0,))

            pix_grid = np.array([[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0],
                                 [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                 [1.0, -1.0], [1.0, 0.0], [1.0, 1.0]])

            geometry = pix.geometry_from_pix_sub_grid(pix_grid, buffer=1e-8)

            assert geometry.x_min == -1.0 - 1e-8
            assert geometry.x_max == 1.0 + 1e-8
            assert geometry.x_pixel_scale == (geometry.x_max - geometry.x_min) / 3
            assert geometry.y_min == -1.0 - 1e-8
            assert geometry.y_max == 1.0 + 1e-8
            assert geometry.y_pixel_scale == (geometry.y_max - geometry.y_min) / 3

        def test__3x3_grid__same_as_above_change_buffer(self):

            pix = pixelization.Rectangular(shape=(3, 3), regularization_coefficients=(1.0,))

            pix_grid = np.array([[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0],
                                 [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                 [1.0, -1.0], [1.0, 0.0], [1.0, 1.0]])

            geometry = pix.geometry_from_pix_sub_grid(pix_grid, buffer=1e-4)

            assert geometry.x_min == -1.0 - 1e-4
            assert geometry.x_max == 1.0 + 1e-4
            assert geometry.x_pixel_scale == (geometry.x_max - geometry.x_min) / 3
            assert geometry.y_pixel_scale == (geometry.y_max - geometry.y_min) / 3
            assert geometry.y_min == -1.0 - 1e-4
            assert geometry.y_max == 1.0 + 1e-4

        def test__5x4_grid__buffer_is_small(self):

            pix = pixelization.Rectangular(shape=(5, 4), regularization_coefficients=(1.0,))

            pix_grid = np.array([[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0],
                                 [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                 [1.0, -1.0], [1.0, 0.0], [1.0, 1.0]])

            geometry = pix.geometry_from_pix_sub_grid(pix_grid, buffer=1e-8)

            assert geometry.x_min == -1.0 - 1e-8
            assert geometry.x_max == 1.0 + 1e-8
            assert geometry.x_pixel_scale == (geometry.x_max - geometry.x_min) / 5
            assert geometry.y_min == -1.0 - 1e-8
            assert geometry.y_max == 1.0 + 1e-8
            assert geometry.y_pixel_scale == (geometry.y_max - geometry.y_min) / 4


        def test__3x3_grid__larger_range_of_grid(self):

            pix = pixelization.Rectangular(shape=(3, 3), regularization_coefficients=(1.0,))

            pix_grid = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]])

            geometry = pix.geometry_from_pix_sub_grid(pix_grid, buffer=1e-8)

            assert geometry.x_min == 1.0 - 1e-8
            assert geometry.x_max == 7.0 + 1e-8
            assert geometry.x_pixel_scale == pytest.approx((geometry.x_max - geometry.x_min) / 3, 1e-4)
            assert geometry.y_min == 2.0 - 1e-8
            assert geometry.y_max == 8.0 + 1e-8
            assert geometry.y_pixel_scale == pytest.approx((geometry.y_max - geometry.y_min) / 3, 1e-4)

    class TestComputeSourceNeighbors:

        def test__compute_pix_neighbors__3x3_grid(self):
            # |0|1|2|
            # |3|4|5|
            # |6|7|8|

            pix = pixelization.Rectangular(shape=(3, 3), regularization_coefficients=(1.0,))

            pix_neighbors = pix.neighbors_from_pixelization()

            assert pix_neighbors[0] == [1, 3]
            assert pix_neighbors[1] == [0, 2, 4]
            assert pix_neighbors[2] == [1, 5]
            assert pix_neighbors[3] == [0, 4, 6]
            assert pix_neighbors[4] == [1, 3, 5, 7]
            assert pix_neighbors[5] == [2, 4, 8]
            assert pix_neighbors[6] == [3, 7]
            assert pix_neighbors[7] == [4, 6, 8]
            assert pix_neighbors[8] == [5, 7]

        def test__compute_pix_neighbors__3x4_grid(self):
            # |0|1| 2| 3|
            # |4|5| 6| 7|
            # |8|9|10|11|

            pix = pixelization.Rectangular(shape=(3, 4), regularization_coefficients=(1.0,))

            pix_neighbors = pix.neighbors_from_pixelization()

            assert pix_neighbors[0] == [1, 4]
            assert pix_neighbors[1] == [0, 2, 5]
            assert pix_neighbors[2] == [1, 3, 6]
            assert pix_neighbors[3] == [2, 7]
            assert pix_neighbors[4] == [0, 5, 8]
            assert pix_neighbors[5] == [1, 4, 6, 9]
            assert pix_neighbors[6] == [2, 5, 7, 10]
            assert pix_neighbors[7] == [3, 6, 11]
            assert pix_neighbors[8] == [4, 9]
            assert pix_neighbors[9] == [5, 8, 10]
            assert pix_neighbors[10] == [6, 9, 11]
            assert pix_neighbors[11] == [7, 10]

        def test__compute_pix_neighbors__4x3_grid(self):
            # |0| 1| 2|
            # |3| 4| 5|
            # |6| 7| 8|
            # |9|10|11|

            pix = pixelization.Rectangular(shape=(4, 3), regularization_coefficients=(1.0,))

            pix_neighbors = pix.neighbors_from_pixelization()

            assert pix_neighbors[0] == [1, 3]
            assert pix_neighbors[1] == [0, 2, 4]
            assert pix_neighbors[2] == [1, 5]
            assert pix_neighbors[3] == [0, 4, 6]
            assert pix_neighbors[4] == [1, 3, 5, 7]
            assert pix_neighbors[5] == [2, 4, 8]
            assert pix_neighbors[6] == [3, 7, 9]
            assert pix_neighbors[7] == [4, 6, 8, 10]
            assert pix_neighbors[8] == [5, 7, 11]
            assert pix_neighbors[9] == [6, 10]
            assert pix_neighbors[10] == [7, 9, 11]
            assert pix_neighbors[11] == [8, 10]

        def test__compute_pix_neighbors__4x4_grid(self):
            # |0 | 1| 2| 3|
            # |4 | 5| 6| 7|
            # |8 | 9|10|11|
            # |12|13|14|15|

            pix = pixelization.Rectangular(shape=(4, 4), regularization_coefficients=(1.0,))

            pix_neighbors = pix.neighbors_from_pixelization()

            assert pix_neighbors[0] == [1, 4]
            assert pix_neighbors[1] == [0, 2, 5]
            assert pix_neighbors[2] == [1, 3, 6]
            assert pix_neighbors[3] == [2, 7]
            assert pix_neighbors[4] == [0, 5, 8]
            assert pix_neighbors[5] == [1, 4, 6, 9]
            assert pix_neighbors[6] == [2, 5, 7, 10]
            assert pix_neighbors[7] == [3, 6, 11]
            assert pix_neighbors[8] == [4, 9, 12]
            assert pix_neighbors[9] == [5, 8, 10, 13]
            assert pix_neighbors[10] == [6, 9, 11, 14]
            assert pix_neighbors[11] == [7, 10, 15]
            assert pix_neighbors[12] == [8, 13]
            assert pix_neighbors[13] == [9, 12, 14]
            assert pix_neighbors[14] == [10, 13, 15]
            assert pix_neighbors[15] == [11, 14]

    class TestComputeImageToSource:

        def test__3x3_grid_of_pix_grid__1_coordinate_per_square_pix_pixel__in_centre_of_pixels(self):
            #   _ _ _
            #  |_|_|_| Boundaries for pixels x = 0 and y = 0  -1.0 to -(1/3)
            #  |_|_|_| Boundaries for pixels x = 1 and y = 1 - (1/3) to (1/3)
            #  |_|_|_| Boundaries for pixels x = 2 and y = 2 - (1/3)" to 1.0"

            pix_grid = np.array([[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0],
                                 [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                 [1.0, -1.0], [1.0, 0.0], [1.0, 1.0]])

            pix = pixelization.Rectangular(shape=(3, 3), regularization_coefficients=(1.0,))

            image_to_pix = pix.grid_to_pix_from_grid_jit(pix_grid, pix.geometry_from_pix_sub_grid(pix_grid))

            assert (image_to_pix == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])).all()

        def test__3x3_grid_of_pix_grid__1_coordinate_per_square_pix_pixel__near_edges_of_pixels(self):
            #   _ _ _
            #  |_|_|_| Boundaries for pixels x = 0 and y = 0  -1.0 to -(1/3)
            #  |_|_|_| Boundaries for pixels x = 1 and y = 1 - (1/3) to (1/3)
            #  |_|_|_| Boundaries for pixels x = 2 and y = 2 - (1/3)" to 1.0"

            pix_grid = np.array([[-0.34, -0.34], [-0.34, 0.325], [-1.0, 1.0],
                                 [-0.32, -1.0], [-0.32, 0.32], [0.0, 1.0],
                                 [1.0, -1.0], [1.0, 0.0], [1.0, 1.0]])

            pix = pixelization.Rectangular(shape=(3, 3), regularization_coefficients=(1.0,))

            image_to_pix = pix.grid_to_pix_from_grid_jit(pix_grid, pix.geometry_from_pix_sub_grid(pix_grid))

            assert (image_to_pix == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])).all()

        def test__3x3_grid_of_pix_grid__add_multiple_grid_to_1_pix_pixel(self):
            #                  _ _ _
            # -1.0 to -(1/3)  |_|_|_|
            # -(1/3) to (1/3) |_|_|_|
            #  (1/3) to 1.0   |_|_|_|

            pix_grid = np.array([[-1.0, -1.0], [0.0, 0.0], [-1.0, 1.0],
                                 [0.0, 0.0], [0.0, 0.0], [0.0, 0.0],
                                 [1.0, -1.0], [0.0, 0.0], [1.0, 1.0]])

            pix = pixelization.Rectangular(shape=(3, 3), regularization_coefficients=(1.0,))

            image_to_pix = pix.grid_to_pix_from_grid_jit(pix_grid, pix.geometry_from_pix_sub_grid(pix_grid))

            assert (image_to_pix == np.array([0, 4, 2, 4, 4, 4, 6, 4, 8])).all()

        def test__4x3_grid_of_pix_grid__1_coordinate_in_each_pixel(self):
            #   _ _ _
            #  |_|_|_|
            #  |_|_|_|
            #  |_|_|_|
            #  |_|_|_|

            # Boundaries for column pixel 0 -1.0 to -(1/3)
            # Boundaries for column pixel 1 -(1/3) to (1/3)
            # Boundaries for column pixel 2  (1/3) to 1.0

            # Bounadries for row pixel 0 -1.0 to -0.5
            # Bounadries for row pixel 1 -0.5 to 0.0
            # Bounadries for row pixel 2  0.0 to 0.5
            # Bounadries for row pixel 3  0.5 to 1.0

            pix_grid = np.array([[-1.0, -1.0], [-1.0, -0.32], [-1.0, 0.34], [-0.49, -1.0],
                                 [0.01, 0.34], [1.0, 1.0]])

            pix = pixelization.Rectangular(shape=(4, 3), regularization_coefficients=(1.0,))

            image_to_pix = pix.grid_to_pix_from_grid_jit(pix_grid, pix.geometry_from_pix_sub_grid(pix_grid))

            assert (image_to_pix == np.array([0, 1, 2, 3, 8, 11])).all()

        def test__3x4_grid_of_pix_grid__1_coordinate_in_each_pixel(self):
            #   _ _ _ _
            #  |_|_|_|_|
            #  |_|_|_|_|
            #  |_|_|_|_|

            # Boundaries for row pixel 0 -1.0 to -(1/3)
            # Boundaries for row pixel 1 -(1/3) to (1/3)
            # Boundaries for row pixel 2  (1/3) to 1.0

            # Bounadries for column pixel 0 -1.0 to -0.5
            # Bounadries for column pixel 1 -0.5 to 0.0
            # Bounadries for column pixel 2  0.0 to 0.5
            # Bounadries for column pixel 3  0.5 to 1.0

            pix_grid = np.array([[-1.0, -1.0], [-1.0, -0.49], [-1.0, 0.01], [-0.32, 0.01],
                                 [0.34, -0.01], [1.0, 1.0]])

            pix = pixelization.Rectangular(shape=(3, 4), regularization_coefficients=(1.0,))

            image_to_pix = pix.grid_to_pix_from_grid_jit(pix_grid, pix.geometry_from_pix_sub_grid(pix_grid))

            assert (image_to_pix == np.array([0, 1, 2, 6, 9, 11])).all()

        def test__3x3_grid__change_arcsecond_dimensions_size__grid_adapts_accordingly(self):
            #   _ _ _
            #  |_|_|_| Boundaries for pixels x = 0 and y = 0  -1.5 to -0.5
            #  |_|_|_| Boundaries for pixels x = 1 and y = 1 -0.5 to 0.5
            #  |_|_|_| Boundaries for pixels x = 2 and y = 2  0.5 to 1.5

            pix_grid = np.array([[-1.5, -1.5], [-1.0, 0.0], [-1.0, 0.6], [1.4, 0.0], [1.5, 1.5]])

            pix = pixelization.Rectangular(shape=(3, 3), regularization_coefficients=(1.0,))

            image_to_pix = pix.grid_to_pix_from_grid_jit(pix_grid, pix.geometry_from_pix_sub_grid(pix_grid))

            assert (image_to_pix == np.array([0, 1, 2, 7, 8])).all()

        def test__3x3_grid__change_arcsecond_dimensions__not_symmetric(self):
            #   _ _ _
            #  |_|_|_| Boundaries for pixels x = 0 and y = 0  -1.5 to -0.5
            #  |_|_|_| Boundaries for pixels x = 1 and y = 1 -0.5 to 0.5
            #  |_|_|_| Boundaries for pixels x = 2 and y = 2  0.5 to 1.5

            pix_grid = np.array([[-1.0, -1.5], [-1.0, -0.49], [-0.32, -1.5], [-0.32, 0.51], [1.0, 1.5]])

            pix = pixelization.Rectangular(shape=(3, 3), regularization_coefficients=(1.0,))

            image_to_pix = pix.grid_to_pix_from_grid_jit(pix_grid, pix.geometry_from_pix_sub_grid(pix_grid))

            assert (image_to_pix == np.array([0, 1, 3, 5, 8])).all()

        def test__4x3_grid__change_arcsecond_dimensions__not_symmetric(self):
            #   _ _ _
            #  |_|_|_|
            #  |_|_|_|
            #  |_|_|_|
            #  |_|_|_|

            pix_grid = np.array([[-1.0, -1.5], [-1.0, -0.49], [-0.49, -1.5], [0.6, 0.0], [1.0, 1.5]])

            pix = pixelization.Rectangular(shape=(4, 3), regularization_coefficients=(1.0,))

            image_to_pix = pix.grid_to_pix_from_grid_jit(pix_grid, pix.geometry_from_pix_sub_grid(pix_grid))

            assert (image_to_pix == np.array([0, 1, 3, 10, 11])).all()

        def test__3x4_grid__change_arcsecond_dimensions__not_symmetric(self):
            #   _ _ _ _
            #  |_|_|_|_|
            #  |_|_|_|_|
            #  |_|_|_|_|

            pix_grid = np.array([[-1.0, -1.5], [-1.0, -0.49], [-0.32, -1.5], [0.34, 0.49], [1.0, 1.5]])

            pix = pixelization.Rectangular(shape=(3, 4), regularization_coefficients=(1.0,))

            image_to_pix = pix.grid_to_pix_from_grid_jit(pix_grid, pix.geometry_from_pix_sub_grid(pix_grid))

            assert (image_to_pix == np.array([0, 1, 4, 10, 11])).all()

    class TestComputeInversion:

        def test__5_simple_grid__no_sub_grid__sets_up_correct_reconstructor(self):

            # Source-plane comprises 5 grid, so 5 image pixels traced to the pix-plane.
            pix_grid = np.array([[-1.0, -1.0], [-1.0, 1.0], [0.0, 0.0], [1.0, -1.0], [1.0, 1.0]])
            pix_border = mask.ImageGridBorder(arr=np.array([0, 1, 3, 4]))

            pix_sub_grid = np.array([[-1.0, -1.0], [-1.0, 1.0], [0.0, 0.0], [1.0, -1.0], [1.0, 1.0]])
            pix_sub_border = mask.SubGridBorder(arr=np.array([0, 1, 3, 4]), sub_grid_size=1)

            sub_to_image = np.array([0, 1, 2, 3, 4])

            grids = MockGridCollection(image=pix_grid,
                                       sub=MockSubGridCoords(pix_sub_grid, sub_to_image,
                                                                   sub_grid_size=1))

            borders = MockBorderCollection(image=pix_border, sub=pix_sub_border)

            # There is no sub-grid, so our sub_grid are just the image grid (note the NumPy weighted_data structure
            # ensures this has no sub-gridding)

            pix = pixelization.RectangularRegConst(shape=(3, 3), regularization_coefficients=(1.0,))

            reconstructor = pix.reconstructor_from_pix_grids(grids, borders)

            assert (reconstructor.mapping == np.array([[1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0]])).all()

            assert (reconstructor.regularization == np.array([[2.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                                             [-1.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                                                             [0.0, -1.0, 2.00000001, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
                                                             [-1.0, 0.0, 0.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0],
                                                             [0.0, -1.0, 0.0, -1.0, 4.00000001, -1.0, 0.0, -1.0, 0.0],
                                                             [0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, 0.0, 0.0, -1.0],
                                                             [0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 2.00000001, -1.0, 0.0],
                                                             [0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, -1.0],
                                                             [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0,
                                                              2.00000001]])).all()

            assert (reconstructor.image_to_pix == np.array([0, 2, 4, 6, 8])).all()
            assert (reconstructor.sub_to_pix == np.array([0, 2, 4, 6, 8])).all()

        def test__15_grid__no_sub_grid__sets_up_correct_reconstructor(self):

            # Source-plane comprises 15 grid, so 15 image pixels traced to the pix-plane.

            pix_grid = np.array([[-0.9, -0.9], [-1.0, -1.0], [-1.1, -1.1],
                                 [-0.9, 0.9], [-1.0, 1.0], [-1.1, 1.1],
                                 [-0.01, 0.01], [0.0, 0.0], [0.01, 0.01],
                                 [0.9, -0.9], [1.0, -1.0], [1.1, -1.1],
                                 [0.9, 0.9], [1.0, 1.0], [1.1, 1.1]])

            pix_border = mask.ImageGridBorder(arr=np.array([2, 5, 11, 14]))

            # There is no sub-grid, so our sub_grid are just the image grid (note the NumPy weighted_data structure
            # ensures this has no sub-gridding)
            pix_sub_grid = np.array([[-0.9, -0.9], [-1.0, -1.0], [-1.1, -1.1],
                                     [-0.9, 0.9], [-1.0, 1.0], [-1.1, 1.1],
                                     [-0.01, 0.01], [0.0, 0.0], [0.01, 0.01],
                                     [0.9, -0.9], [1.0, -1.0], [1.1, -1.1],
                                     [0.9, 0.9], [1.0, 1.0], [1.1, 1.1]])

            pix_sub_border = mask.SubGridBorder(arr=np.array([2, 5, 11, 14]))

            sub_to_image = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14])

            grids = MockGridCollection(image=pix_grid,
                                       sub=MockSubGridCoords(pix_sub_grid, sub_to_image,
                                                                   sub_grid_size=1))

            borders = MockBorderCollection(image=pix_border, sub=pix_sub_border)

            pix = pixelization.RectangularRegConst(shape=(3, 3), regularization_coefficients=(1.0,))

            reconstructor = pix.reconstructor_from_pix_grids(grids, borders)

            assert (reconstructor.mapping == np.array([[1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                                      [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                                      [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
                                                      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
                                                      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0]])).all()

            assert (reconstructor.regularization == np.array([[2.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                                             [-1.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                                                             [0.0, -1.0, 2.00000001, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
                                                             [-1.0, 0.0, 0.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0],
                                                             [0.0, -1.0, 0.0, -1.0, 4.00000001, -1.0, 0.0, -1.0, 0.0],
                                                             [0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, 0.0, 0.0, -1.0],
                                                             [0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 2.00000001, -1.0, 0.0],
                                                             [0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, -1.0],
                                                             [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0,
                                                              2.00000001]])).all()

            assert (reconstructor.image_to_pix == np.array([0, 0, 0, 2, 2, 2, 4, 4, 4, 6, 6, 6, 8, 8, 8])).all()
            assert (reconstructor.sub_to_pix == np.array([0, 0, 0, 2, 2, 2, 4, 4, 4, 6, 6, 6, 8, 8, 8])).all()

        def test__5_simple_grid__include_sub_grid__sets_up_correct_reconstructor(self):

            # Source-plane comprises 5 grid, so 5 image pixels traced to the pix-plane.
            pix_grid = np.array([[-1.0, -1.0], [-1.0, 1.0], [0.0, 0.0], [1.0, -1.0], [1.0, 1.0]])
            pix_border = mask.ImageGridBorder(arr=np.array([0, 1, 3, 4]))
            # Assume a 2x2 sub-grid, so each of our 5 image-pixels are split into 4.
            # The grid below is unphysical in that the (0.0, 0.0) terms on the end of each sub-grid probably couldn't
            # happen for a real lensing calculation. This is to make a mapping matrix which explicitly tests the 
            # sub-grid.
            pix_sub_grid = np.array([[-1.0, -1.0], [-1.0, -1.0], [-1.0, -1.0], [0.0, 0.0],
                                     [-1.0, 1.0], [-1.0, 1.0], [-1.0, 1.0], [0.0, 0.0],
                                     [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0],
                                     [1.0, -1.0], [1.0, -1.0], [1.0, -1.0], [0.0, 0.0],
                                     [1.0, 1.0], [1.0, 1.0], [1.0, 1.0], [0.0, 0.0]])

            sub_to_image = np.array([0, 0, 0, 2, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 2, 4, 4, 4, 2])
            pix_sub_border = mask.SubGridBorder(arr=np.array([0, 1, 2, 4, 5, 6, 12, 13, 14, 16, 17, 18]))

            grids = MockGridCollection(image=pix_grid,
                                       sub=MockSubGridCoords(pix_sub_grid, sub_to_image,
                                                                   sub_grid_size=2))

            borders = MockBorderCollection(image=pix_border, sub=pix_sub_border)

            pix = pixelization.RectangularRegConst(shape=(3, 3), regularization_coefficients=(1.0,))

            reconstructor = pix.reconstructor_from_pix_grids(grids, borders)

            assert (reconstructor.mapping == np.array([[0.75, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.75, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.75, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.75]])).all()

            assert (reconstructor.regularization == np.array([[2.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                                             [-1.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                                                             [0.0, -1.0, 2.00000001, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
                                                             [-1.0, 0.0, 0.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0],
                                                             [0.0, -1.0, 0.0, -1.0, 4.00000001, -1.0, 0.0, -1.0, 0.0],
                                                             [0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, 0.0, 0.0, -1.0],
                                                             [0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 2.00000001, -1.0, 0.0],
                                                             [0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, -1.0],
                                                             [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0,
                                                              2.00000001]])).all()

            assert (reconstructor.image_to_pix == np.array([0, 2, 4, 6, 8])).all()
            assert (reconstructor.sub_to_pix == np.array(
                [0, 0, 0, 4, 2, 2, 2, 4, 4, 4, 4, 4, 6, 6, 6, 4, 8, 8, 8, 4])).all()

        def test__same_as_above_but_grid_requires_border_relocation(self):

            # Source-plane comprises 5 grid, so 5 image pixels traced to the pix-plane.
            pix_grid = np.array([[-1.0, -1.0], [-1.0, 1.0], [0.0, 0.0], [1.0, -1.0], [1.0, 1.0]])
            pix_border = mask.ImageGridBorder(arr=np.array([0, 1, 3, 4]))
            # Assume a 2x2 sub-grid, so each of our 5 image-pixels are split into 4.
            # The grid below is unphysical in that the (0.0, 0.0) terms on the end of each sub-grid probably couldn't
            # happen for a real lensing calculation. This is to make a mapping matrix which explicitly tests the
            # sub-grid.
            pix_sub_grid = np.array([[-1.0, -1.0], [-2.0, -2.0], [-2.0, -2.0], [0.0, 0.0],
                                     [-1.0, 1.0], [-2.0, 2.0], [-2.0, 2.0], [0.0, 0.0],
                                     [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0],
                                     [1.0, -1.0], [2.0, -2.0], [2.0, -2.0], [0.0, 0.0],
                                     [1.0, 1.0], [2.0, 2.0], [2.0, 2.0], [0.0, 0.0]])


            pix_sub_border = mask.SubGridBorder(arr=np.array([0, 4, 12, 16]))

            sub_to_image = np.array([0, 0, 0, 2, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 2, 4, 4, 4, 2])

            grids = MockGridCollection(image=pix_grid,
                                       sub=MockSubGridCoords(pix_sub_grid, sub_to_image,
                                                                   sub_grid_size=2))

            borders = MockBorderCollection(image=pix_border, sub=pix_sub_border)

            pix = pixelization.RectangularRegConst(shape=(3, 3), regularization_coefficients=(1.0,))

            reconstructor = pix.reconstructor_from_pix_grids(grids, borders)

            assert (reconstructor.mapping == np.array([[0.75, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.75, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.75, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.75]])).all()

            assert (reconstructor.regularization == np.array([[2.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                                             [-1.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                                                             [0.0, -1.0, 2.00000001, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
                                                             [-1.0, 0.0, 0.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0],
                                                             [0.0, -1.0, 0.0, -1.0, 4.00000001, -1.0, 0.0, -1.0, 0.0],
                                                             [0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, 0.0, 0.0, -1.0],
                                                             [0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 2.00000001, -1.0, 0.0],
                                                             [0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, -1.0],
                                                             [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0,
                                                              2.00000001]])).all()

            assert (reconstructor.image_to_pix == np.array([0, 2, 4, 6, 8])).all()
            assert (reconstructor.sub_to_pix == np.array(
                [0, 0, 0, 4, 2, 2, 2, 4, 4, 4, 4, 4, 6, 6, 6, 4, 8, 8, 8, 4])).all()


class TestVoronoiPixelization:

    class TestComputeVoronoi:

        def test__points_in_x_cross_shape__sets_up_diamond_voronoi_vertices(self):
            # 5 points in the shape of the face of a 5 on a die - makes a diamond Voronoi diagram

            points = np.array([[-1.0, 1.0], [1.0, 1.0],
                               [0.0, 0.0],
                               [-1.0, -1.0], [1.0, -1.0]])

            pix = pixelization.Voronoi(pixels=5, regularization_coefficients=1.0)
            voronoi = pix.voronoi_from_cluster_grid(points)

            voronoi.vertices = list(map(lambda x: list(x), voronoi.vertices))

            assert [0, 1.] in voronoi.vertices
            assert [-1., 0.] in voronoi.vertices
            assert [1., 0.] in voronoi.vertices
            assert [0., -1.] in voronoi.vertices

        def test__9_points_in_square___sets_up_square_of_voronoi_vertices(self):
            # 9 points in a square - makes a square (this is the example int he scipy documentaiton page)

            points = np.array([[0.0, 0.0], [1.0, 0.0], [2.0, 0.0],
                               [0.0, 1.0], [1.0, 1.0], [2.0, 1.0],
                               [0.0, 2.0], [1.0, 2.0], [2.0, 2.0]])

            pix = pixelization.Voronoi(pixels=9, regularization_coefficients=1.0)
            voronoi = pix.voronoi_from_cluster_grid(points)

            # ridge points is a numpy array for speed, but convert to list for the comparisons below so we can use in
            # to look for each list

            voronoi.vertices = list(map(lambda x: list(x), voronoi.vertices))

            assert [0.5, 1.5] in voronoi.vertices
            assert [1.5, 0.5] in voronoi.vertices
            assert [0.5, 0.5] in voronoi.vertices
            assert [1.5, 1.5] in voronoi.vertices

        def test__points_in_x_cross_shape__sets_up_pairs_of_voronoi_cells(self):
            # 5 points in the shape of the face of a 5 on a die - makes a diamond Voronoi diagram

            points = np.array([[-1.0, 1.0], [1.0, 1.0],
                               [0.0, 0.0],
                               [-1.0, -1.0], [1.0, -1.0]])

            pix = pixelization.Voronoi(pixels=5, regularization_coefficients=1.0)
            voronoi = pix.voronoi_from_cluster_grid(points)

            # ridge points is a numpy array for speed, but convert to list for the comparisons below so we can use in
            # to look for each list

            voronoi.ridge_points = list(map(lambda x: list(x), voronoi.ridge_points))

            assert len(voronoi.ridge_points) == 8

            assert [2, 0] in voronoi.ridge_points or [0, 2] in voronoi.ridge_points
            assert [2, 1] in voronoi.ridge_points or [1, 2] in voronoi.ridge_points
            assert [2, 3] in voronoi.ridge_points or [3, 2] in voronoi.ridge_points
            assert [2, 4] in voronoi.ridge_points or [4, 2] in voronoi.ridge_points
            assert [0, 1] in voronoi.ridge_points or [1, 0] in voronoi.ridge_points
            assert [0.3] in voronoi.ridge_points or [3, 0] in voronoi.ridge_points
            assert [3, 4] in voronoi.ridge_points or [4, 3] in voronoi.ridge_points
            assert [4, 1] in voronoi.ridge_points or [1, 4] in voronoi.ridge_points

        def test__9_points_in_square___sets_up_pairs_of_voronoi_cells(self):
            # 9 points in a square - makes a square (this is the example int he scipy documentaiton page)

            points = np.array([[0.0, 0.0], [1.0, 0.0], [2.0, 0.0],
                               [0.0, 1.0], [1.0, 1.0], [2.0, 1.0],
                               [0.0, 2.0], [1.0, 2.0], [2.0, 2.0]])

            pix = pixelization.Voronoi(pixels=9, regularization_coefficients=1.0)
            voronoi = pix.voronoi_from_cluster_grid(points)

            # ridge points is a numpy array for speed, but convert to list for the comparisons below so we can use in
            # to look for each list

            voronoi.ridge_points = list(map(lambda x: list(x), voronoi.ridge_points))

            assert len(voronoi.ridge_points) == 12

            assert [0, 1] in voronoi.ridge_points or [1, 0] in voronoi.ridge_points
            assert [1, 2] in voronoi.ridge_points or [2, 1] in voronoi.ridge_points
            assert [3, 4] in voronoi.ridge_points or [4, 3] in voronoi.ridge_points
            assert [4, 5] in voronoi.ridge_points or [5, 4] in voronoi.ridge_points
            assert [6, 7] in voronoi.ridge_points or [7, 6] in voronoi.ridge_points
            assert [7, 8] in voronoi.ridge_points or [8, 7] in voronoi.ridge_points

            assert [0, 3] in voronoi.ridge_points or [3, 0] in voronoi.ridge_points
            assert [1, 4] in voronoi.ridge_points or [4, 1] in voronoi.ridge_points
            assert [4, 7] in voronoi.ridge_points or [7, 4] in voronoi.ridge_points
            assert [2, 5] in voronoi.ridge_points or [5, 2] in voronoi.ridge_points
            assert [5, 8] in voronoi.ridge_points or [8, 5] in voronoi.ridge_points
            assert [3, 6] in voronoi.ridge_points or [6, 3] in voronoi.ridge_points

    class TestComputeNeighbors:

        def test__points_in_x_cross_shape__neighbors_of_each_pix_pixel_correct(self):
            # 5 points in the shape of the face of a 5 on a die - makes a diamond Voronoi diagram

            points = np.array([[-1.0, 1.0], [1.0, 1.0],
                               [0.0, 0.0],
                               [-1.0, -1.0], [1.0, -1.0]])

            pix = pixelization.Voronoi(pixels=5, regularization_coefficients=1.0)
            voronoi = pix.voronoi_from_cluster_grid(points)
            neighbors = pix.neighbors_from_pixelization(voronoi.ridge_points)

            assert set(neighbors[0]) == {2, 1, 3}
            assert set(neighbors[1]) == {2, 0, 4}
            assert set(neighbors[2]) == {0, 1, 3, 4}
            assert set(neighbors[3]) == {2, 0, 4}
            assert set(neighbors[4]) == {2, 1, 3}

        def test__9_points_in_square___neighbors_of_each_pix_pixel_correct(self):
            # 9 points in a square - makes a square (this is the example int he scipy documentaiton page)

            points = np.array([[0.0, 0.0], [1.0, 0.0], [2.0, 0.0],
                               [0.0, 1.0], [1.0, 1.0], [2.0, 1.0],
                               [0.0, 2.0], [1.0, 2.0], [2.0, 2.0]])

            pix = pixelization.Voronoi(pixels=9, regularization_coefficients=1.0)
            voronoi = pix.voronoi_from_cluster_grid(points)
            neighbors = pix.neighbors_from_pixelization(voronoi.ridge_points)

            assert set(neighbors[0]) == {1, 3}
            assert set(neighbors[1]) == {0, 2, 4}
            assert set(neighbors[2]) == {1, 5}
            assert set(neighbors[3]) == {0, 4, 6}
            assert set(neighbors[4]) == {1, 3, 5, 7}
            assert set(neighbors[5]) == {2, 4, 8}
            assert set(neighbors[6]) == {3, 7}
            assert set(neighbors[7]) == {4, 6, 8}
            assert set(neighbors[8]) == {5, 7}

    class TestImageToSourceViaNearestNeighborsForTesting:

        def test__grid_to_pix_pixels_via_nearest_neighbour__case1__correct_pairs(self):
            pix_pixels = np.array([[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0]])
            image_grid = np.array([[1.1, 1.1], [-1.1, 1.1], [-1.1, -1.1], [1.1, -1.1]])

            image_to_pix = grid_to_pix_pixels_via_nearest_neighbour(image_grid, pix_pixels)

            assert image_to_pix[0] == 0
            assert image_to_pix[1] == 1
            assert image_to_pix[2] == 2
            assert image_to_pix[3] == 3

        def test__grid_to_pix_pixels_via_nearest_neighbour___case2__correct_pairs(self):
            pix_pixels = np.array([[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0]])
            image_grid = np.array([[1.1, 1.1], [-1.1, 1.1], [-1.1, -1.1], [1.1, -1.1],
                                   [0.9, -0.9], [-0.9, -0.9], [-0.9, 0.9], [0.9, 0.9]])

            image_to_pix = grid_to_pix_pixels_via_nearest_neighbour(image_grid, pix_pixels)

            assert image_to_pix[0] == 0
            assert image_to_pix[1] == 1
            assert image_to_pix[2] == 2
            assert image_to_pix[3] == 3
            assert image_to_pix[4] == 3
            assert image_to_pix[5] == 2
            assert image_to_pix[6] == 1
            assert image_to_pix[7] == 0

        def test__grid_to_pix_pixels_via_nearest_neighbour___case3__correct_pairs(self):
            pix_pixels = np.array([[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0], [0.0, 0.0], [2.0, 2.0]])
            image_grid = np.array([[0.1, 0.1], [-0.1, -0.1], [0.49, 0.49],
                                   [0.51, 0.51], [1.01, 1.01], [1.51, 1.51]])

            image_to_pix = grid_to_pix_pixels_via_nearest_neighbour(image_grid, pix_pixels)

            assert image_to_pix[0] == 4
            assert image_to_pix[1] == 4
            assert image_to_pix[2] == 4
            assert image_to_pix[3] == 0
            assert image_to_pix[4] == 0
            assert image_to_pix[5] == 5

    class TestSubToSourceViaNearestNeighborsForTesting:

        def test__grid_to_pix_pixels_via_nearest_neighbour__case1__correct_pairs(self):
            pix_pixels = np.array([[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0]])
            sub_grid = np.array([[1.1, 1.1], [-1.1, 1.1], [-1.1, -1.1], [1.1, -1.1]])

            sub_to_pix = grid_to_pix_pixels_via_nearest_neighbour(sub_grid, pix_pixels)

            assert sub_to_pix[0] == 0
            assert sub_to_pix[1] == 1
            assert sub_to_pix[2] == 2
            assert sub_to_pix[3] == 3

        def test__grid_to_pix_pixels_via_nearest_neighbour___case2__correct_pairs(self):
            pix_pixels = np.array([[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0]])
            sub_grid = np.array([[1.1, 1.1], [-1.1, 1.1], [-1.1, -1.1], [1.1, -1.1],
                                 [0.9, -0.9], [-0.9, -0.9], [-0.9, 0.9], [0.9, 0.9]])

            sub_to_pix = grid_to_pix_pixels_via_nearest_neighbour(sub_grid, pix_pixels)

            assert sub_to_pix[0] == 0
            assert sub_to_pix[1] == 1
            assert sub_to_pix[2] == 2
            assert sub_to_pix[3] == 3
            assert sub_to_pix[4] == 3
            assert sub_to_pix[5] == 2
            assert sub_to_pix[6] == 1
            assert sub_to_pix[7] == 0

        def test__grid_to_pix_pixels_via_nearest_neighbour___case3__correct_pairs(self):
            pix_pixels = np.array([[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0], [0.0, 0.0], [2.0, 2.0]])
            sub_grid = np.array([[0.1, 0.1], [-0.1, -0.1], [0.49, 0.49],
                                 [0.51, 0.51], [1.01, 1.01], [1.51, 1.51]])

            sub_to_pix = grid_to_pix_pixels_via_nearest_neighbour(sub_grid, pix_pixels)

            assert sub_to_pix[0] == 4
            assert sub_to_pix[1] == 4
            assert sub_to_pix[2] == 4
            assert sub_to_pix[3] == 0
            assert sub_to_pix[4] == 0
            assert sub_to_pix[5] == 5

    class TestImageToSource:

        def test__image_grid_to_pix_pixels_via_cluster_pairs__grid_of_pix_pixels__correct_pairs(self):
            pix_centers = np.array([[-1.0, -1.0], [-0.9, 0.9],
                                    [1.0, -1.1], [1.2, 1.2]])

            pix_grid = np.array([[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0],
                                 [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                 [1.0, -1.0], [1.0, 0.0], [1.0, 1.0]])

            image_to_pix = np.array([0, 1, 1, 0, 1, 1, 2, 2, 3])

            image_to_pix_via_nearest_neighbour = grid_to_pix_pixels_via_nearest_neighbour(pix_grid,
                                                                                          pix_centers)

            image_to_cluster = np.array([0, 0, 1, 0, 0, 1, 2, 2, 3])
            cluster_to_image = np.array([0, 2, 6, 8])
            cluster_to_pix = np.array([0, 1, 2, 3])
            cluster_mask = MockSparseMask(sparse_to_image=cluster_to_image, image_to_sparse=image_to_cluster)

            sub_to_image = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
            grids = MockGridCollection(image=pix_grid,
                                       sub=MockSubGridCoords(np.array([]), sub_to_image,
                                                                   sub_grid_size=1))

            pix = pixelization.Voronoi(pixels=6, regularization_coefficients=1.0)
            voronoi = pix.voronoi_from_cluster_grid(pix_centers)
            pix_neighbors = pix.neighbors_from_pixelization(voronoi.ridge_points)

            image_to_pix_via_pairs = pix.image_to_pix_from_pixelization(grids, pix_centers, pix_neighbors,
                                                                        cluster_to_pix, cluster_mask)

            assert (image_to_pix_via_pairs == image_to_pix).all()
            assert (image_to_pix_via_pairs == image_to_pix_via_nearest_neighbour).all()

    class TestComputeSubToSource:

        def test__sub_grid_to_pix_pixels_via_cluster_pairs__grid_of_pix_pixels__correct_pairs(self):
            pix_centers = np.array([[0.1, 0.1], [1.1, 0.1], [2.1, 0.1],
                                    [0.1, 1.1], [1.1, 1.1], [2.1, 1.1]])
            pix_sub_grid = np.array([[0.05, 0.15], [0.15, 0.15], [0.05, 0.05], [0.15, 0.05],
                                     [1.05, 0.15], [1.15, 0.15], [1.05, 0.05], [1.15, 0.05],
                                     [2.05, 0.15], [2.15, 0.15], [2.05, 0.05], [2.15, 0.05],
                                     [0.05, 1.15], [0.15, 1.15], [0.05, 1.05], [0.15, 1.05],
                                     [1.05, 1.15], [1.15, 1.15], [1.05, 1.05], [1.15, 1.05],
                                     [2.05, 1.15], [2.15, 1.15], [2.05, 1.05], [2.15, 1.05]])

            sub_to_pix_via_nearest_neighbour = grid_to_pix_pixels_via_nearest_neighbour(pix_sub_grid,
                                                                                        pix_centers)

            image_to_cluster = np.array([0, 0, 1, 0, 0, 1, 2, 2, 3])
            cluster_to_image = np.array([0, 2, 6, 8])
            cluster_to_pix = np.array([0, 1, 2, 3])
            cluster_mask = MockSparseMask(sparse_to_image=cluster_to_image, image_to_sparse=image_to_cluster)

            sub_to_image = np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5])
            grids = MockGridCollection(image=np.array([]),
                                       sub=MockSubGridCoords(pix_sub_grid, sub_to_image,
                                                                   sub_grid_size=1))

            pix = pixelization.Voronoi(pixels=6, regularization_coefficients=1.0)
            voronoi = pix.voronoi_from_cluster_grid(pix_centers)
            pix_neighbors = pix.neighbors_from_pixelization(voronoi.ridge_points)

            sub_to_pix_via_pairs = pix.sub_to_pix_from_pixelization(grids, pix_centers, pix_neighbors,
                                                                    cluster_to_pix, cluster_mask)

            assert (sub_to_pix_via_nearest_neighbour == sub_to_pix_via_pairs).all()


class TestClusterRegConst:
    
    class TestComputeInversion:

        def test__5_simple_grid__no_sub_grid__sets_up_correct_reconstructor(self):

            pix_grid = np.array([[1.0, 1.0], [-1.0, 1.0], [0.0, 0.0], [1.0, -1.0], [-1.0, -1.0]])
            pix_border = mask.ImageGridBorder(arr=np.array([0, 1, 3, 4]))

            pix_sub_grid = np.array([[1.0, 1.0], [-1.0, 1.0], [0.0, 0.0], [1.0, -1.0], [-1.0, -1.0]])
            pix_sub_border = mask.SubGridBorder(arr=np.array([0, 1, 3, 4]), sub_grid_size=1)

            sub_to_image = np.array([0, 1, 2, 3, 4])

            cluster_to_image = np.array([0, 1, 2, 3, 4])
            image_to_cluster = np.array([0, 1, 2, 3, 4])
            cluster_mask = MockSparseMask(sparse_to_image=cluster_to_image, image_to_sparse=image_to_cluster)

            grids = MockGridCollection(image=pix_grid,
                                       sub=MockSubGridCoords(pix_sub_grid, sub_to_image, sub_grid_size=1))

            borders = MockBorderCollection(image=pix_border, sub=pix_sub_border)

            pix = pixelization.ClusterRegConst(pixels=5, regularization_coefficients=(1.0,))

            reconstructor = pix.reconstructor_from_pix_grids(grids, borders, cluster_mask)

            assert (reconstructor.mapping == np.array([[1.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 1.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 1.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 1.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 1.0]])).all()

            assert (reconstructor.regularization == np.array([[3.00000001, -1.0, -1.0, -1.0, 0.0],
                                                             [-1.0, 3.00000001, -1.0, 0.0, -1.0],
                                                             [-1.0, -1.0, 4.00000001, -1.0, -1.0],
                                                             [-1.0, 0.0, -1.0, 3.00000001, -1.0],
                                                             [0.0, -1.0, -1.0, -1.0, 3.00000001]])).all()

            assert (reconstructor.image_to_pix == np.array([0, 1, 2, 3, 4])).all()
            assert (reconstructor.sub_to_pix == np.array([0, 1, 2, 3, 4])).all()

        def test__15_grid__no_sub_grid__sets_up_correct_reconstructor(self):

            cluster_to_image = np.array([1, 4, 7, 10, 13])
            image_to_cluster = np.array([0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4])
            cluster_mask = MockSparseMask(sparse_to_image=cluster_to_image, image_to_sparse=image_to_cluster)

            pix_grid = np.array([[0.9, 0.9], [1.0, 1.0], [1.1, 1.1],
                                 [-0.9, 0.9], [-1.0, 1.0], [-1.1, 1.1],
                                 [-0.01, 0.01], [0.0, 0.0], [0.01, 0.01],
                                 [0.9, -0.9], [1.0, -1.0], [1.1, -1.1],
                                 [-0.9, -0.9], [-1.0, -1.0], [-1.1, -1.1]])

            pix_border = mask.ImageGridBorder(arr=np.array([2, 5, 11, 14]))

            pix_sub_grid = np.array([[0.9, 0.9], [1.0, 1.0], [1.1, 1.1],
                                     [-0.9, 0.9], [-1.0, 1.0], [-1.1, 1.1],
                                     [-0.01, 0.01], [0.0, 0.0], [0.01, 0.01],
                                     [0.9, -0.9], [1.0, -1.0], [1.1, -1.1],
                                     [-0.9, -0.9], [-1.0, -1.0], [-1.1, -1.1]])

            pix_sub_border = mask.SubGridBorder(arr=np.array([2, 5, 11, 14]))

            sub_to_image = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14])
            grids = MockGridCollection(image=pix_grid,
                                       sub=MockSubGridCoords(pix_sub_grid, sub_to_image,
                                                                   sub_grid_size=1))

            borders = MockBorderCollection(image=pix_border, sub=pix_sub_border)


            pix = pixelization.ClusterRegConst(pixels=5, regularization_coefficients=(1.0,))

            reconstructor = pix.reconstructor_from_pix_grids(grids, borders, cluster_mask)

            assert (reconstructor.mapping == np.array([[1.0, 0.0, 0.0, 0.0, 0.0],
                                                      [1.0, 0.0, 0.0, 0.0, 0.0],
                                                      [1.0, 0.0, 0.0, 0.0, 0.0],
                                                      [0.0, 1.0, 0.0, 0.0, 0.0],
                                                      [0.0, 1.0, 0.0, 0.0, 0.0],
                                                      [0.0, 1.0, 0.0, 0.0, 0.0],
                                                      [0.0, 0.0, 1.0, 0.0, 0.0],
                                                      [0.0, 0.0, 1.0, 0.0, 0.0],
                                                      [0.0, 0.0, 1.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.0, 1.0, 0.0],
                                                      [0.0, 0.0, 0.0, 1.0, 0.0],
                                                      [0.0, 0.0, 0.0, 1.0, 0.0],
                                                      [0.0, 0.0, 0.0, 0.0, 1.0],
                                                      [0.0, 0.0, 0.0, 0.0, 1.0],
                                                      [0.0, 0.0, 0.0, 0.0, 1.0]])).all()

            assert (reconstructor.regularization == np.array([[3.00000001, -1.0, -1.0, -1.0, 0.0],
                                                             [-1.0, 3.00000001, -1.0, 0.0, -1.0],
                                                             [-1.0, -1.0, 4.00000001, -1.0, -1.0],
                                                             [-1.0, 0.0, -1.0, 3.00000001, -1.0],
                                                             [0.0, -1.0, -1.0, -1.0, 3.00000001]])).all()

            assert (reconstructor.image_to_pix == np.array([0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4])).all()
            assert (reconstructor.sub_to_pix == np.array([0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4])).all()

        def test__5_simple_grid__include_sub_grid__sets_up_correct_reconstructor(self):

            pix_grid = np.array([[1.0, 1.0], [-1.0, 1.0], [0.0, 0.0], [1.0, -1.0], [-1.0, -1.0]])
            pix_border = mask.ImageGridBorder(arr=np.array([0, 1, 3, 4]))

            cluster_to_image = np.array([0, 1, 2, 3, 4])
            image_to_cluster = np.array([0, 1, 2, 3, 4])
            cluster_mask = MockSparseMask(sparse_to_image=cluster_to_image, image_to_sparse=image_to_cluster)

            pix_sub_grid = np.array([[1.0, 1.0], [1.0, 1.0], [1.0, 1.0], [0.0, 0.0],
                                     [-1.0, 1.0], [-1.0, 1.0], [-1.0, 1.0], [0.0, 0.0],
                                     [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0],
                                     [1.0, -1.0], [1.0, -1.0], [1.0, -1.0], [0.0, 0.0],
                                     [-1.0, -1.0], [-1.0, -1.0], [-1.0, -1.0], [0.0, 0.0]])

            pix_sub_border = mask.SubGridBorder(arr=np.array([0, 1, 2, 4, 5, 6, 12, 13, 14, 16, 17, 18]))

            sub_to_image = np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4])

            grids = MockGridCollection(image=pix_grid,
                                       sub=MockSubGridCoords(pix_sub_grid, sub_to_image,
                                                                   sub_grid_size=2))

            borders = MockBorderCollection(image=pix_border, sub=pix_sub_border)

            pix = pixelization.ClusterRegConst(pixels=5, regularization_coefficients=(1.0,))

            reconstructor = pix.reconstructor_from_pix_grids(grids, borders, cluster_mask)

            assert (reconstructor.mapping == np.array([[0.75, 0.0, 0.25, 0.0, 0.0],
                                                      [0.0, 0.75, 0.25, 0.0, 0.0],
                                                      [0.0, 0.0, 1.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.25, 0.75, 0.0],
                                                      [0.0, 0.0, 0.25, 0.0, 0.75]])).all()

            assert (reconstructor.regularization == np.array([[3.00000001, -1.0, -1.0, -1.0, 0.0],
                                                             [-1.0, 3.00000001, -1.0, 0.0, -1.0],
                                                             [-1.0, -1.0, 4.00000001, -1.0, -1.0],
                                                             [-1.0, 0.0, -1.0, 3.00000001, -1.0],
                                                             [0.0, -1.0, -1.0, -1.0, 3.00000001]])).all()

            assert (reconstructor.image_to_pix == np.array([0, 1, 2, 3, 4])).all()
            assert (reconstructor.sub_to_pix == np.array(
                [0, 0, 0, 2, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 2, 4, 4, 4, 2])).all()

        def test__same_as_above_but_grid_requires_border_relocation(self):

            pix_grid = np.array([[1.0, 1.0], [-1.0, 1.0], [0.0, 0.0], [1.0, -1.0], [-1.0, -1.0]])
            pix_border = mask.ImageGridBorder(arr=np.array([0, 1, 3, 4]))

            cluster_to_image = np.array([0, 1, 2, 3, 4])
            image_to_cluster = np.array([0, 1, 2, 3, 4])
            cluster_mask = MockSparseMask(sparse_to_image=cluster_to_image, image_to_sparse=image_to_cluster)

            pix_sub_grid = np.array([[1.0, 1.0], [2.0, 2.0], [2.0, 2.0], [0.0, 0.0],
                                     [-1.0, 1.0], [-2.0, 2.0], [-2.0, 2.0], [0.0, 0.0],
                                     [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0],
                                     [1.0, -1.0], [2.0, -2.0], [2.0, -2.0], [0.0, 0.0],
                                     [-1.0, -1.0], [-2.0, -2.0], [-2.0, -2.0], [0.0, 0.0]])

            pix_sub_border = mask.SubGridBorder(arr=np.array([0, 4, 12, 16]))

            sub_to_image = np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4])

            grids = MockGridCollection(image=pix_grid,
                                       sub=MockSubGridCoords(pix_sub_grid, sub_to_image,
                                                                   sub_grid_size=2))

            borders = MockBorderCollection(image=pix_border, sub=pix_sub_border)

            pix = pixelization.ClusterRegConst(pixels=5, regularization_coefficients=(1.0,))

            reconstructor = pix.reconstructor_from_pix_grids(grids, borders, cluster_mask)


            assert (reconstructor.mapping == np.array([[0.75, 0.0, 0.25, 0.0, 0.0],
                                                      [0.0, 0.75, 0.25, 0.0, 0.0],
                                                      [0.0, 0.0, 1.0, 0.0, 0.0],
                                                      [0.0, 0.0, 0.25, 0.75, 0.0],
                                                      [0.0, 0.0, 0.25, 0.0, 0.75]])).all()

            assert (reconstructor.regularization == np.array([[3.00000001, -1.0, -1.0, -1.0, 0.0],
                                                             [-1.0, 3.00000001, -1.0, 0.0, -1.0],
                                                             [-1.0, -1.0, 4.00000001, -1.0, -1.0],
                                                             [-1.0, 0.0, -1.0, 3.00000001, -1.0],
                                                             [0.0, -1.0, -1.0, -1.0, 3.00000001]])).all()

            assert (reconstructor.image_to_pix == np.array([0, 1, 2, 3, 4])).all()
            assert (reconstructor.sub_to_pix == np.array(
                [0, 0, 0, 2, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 2, 4, 4, 4, 2])).all()


class TestAmorphousPixelization:

    class TestKMeans:

        def test__simple_points__sets_up_two_clusters(self):
            cluster_grid = np.array([[0.99, 0.99], [1.0, 1.0], [1.01, 1.01],
                                     [1.99, 1.99], [2.0, 2.0], [2.01, 2.01]])

            pix = pixelization.Amorphous(pixels=2)

            pix_centers, pix_to_image = pix.kmeans_cluster(cluster_grid)

            assert [2.0, 2.0] in pix_centers
            assert [1.0, 1.0] in pix_centers

            assert list(pix_to_image).count(0) == 3
            assert list(pix_to_image).count(1) == 3

        def test__simple_points__sets_up_three_clusters(self):
            cluster_grid = np.array([[-0.99, -0.99], [-1.0, -1.0], [-1.01, -1.01],
                                     [0.99, 0.99], [1.0, 1.0], [1.01, 1.01],
                                     [1.99, 1.99], [2.0, 2.0], [2.01, 2.01]])

            pix = pixelization.Amorphous(pixels=3)

            pix_centers, pix_to_image = pix.kmeans_cluster(cluster_grid)

            assert [2.0, 2.0] in pix_centers
            assert [1.0, 1.0] in pix_centers
            assert [-1.0, -1.0] in pix_centers

            assert list(pix_to_image).count(0) == 3
            assert list(pix_to_image).count(1) == 3
            assert list(pix_to_image).count(2) == 3

        def test__simple_points__sets_up_three_clusters_more_points_in_third_cluster(self):
            cluster_grid = np.array([[-0.99, -0.99], [-1.0, -1.0], [-1.01, -1.01],

                                     [0.99, 0.99], [1.0, 1.0], [1.01, 1.01],
                                     [0.99, 0.99], [1.0, 1.0], [1.01, 1.01],

                                     [1.99, 1.99], [2.0, 2.0], [2.01, 2.01],
                                     [1.99, 1.99], [2.0, 2.0], [2.01, 2.01],
                                     [1.99, 1.99], [2.0, 2.0], [2.01, 2.01],
                                     [1.99, 1.99], [2.0, 2.0], [2.01, 2.01]])

            pix = pixelization.Amorphous(pixels=3)

            pix_centers, pix_to_image = pix.kmeans_cluster(cluster_grid)

            pix_centers = list(map(lambda x: pytest.approx(list(x), 1e-3), pix_centers))

            assert [2.0, 2.0] in pix_centers
            assert [1.0, 1.0] in pix_centers
            assert [-1.0, -1.0] in pix_centers

            assert list(pix_to_image).count(0) == 3 or 6 or 12
            assert list(pix_to_image).count(1) == 3 or 6 or 12
            assert list(pix_to_image).count(2) == 3 or 6 or 12

            assert list(pix_to_image).count(0) != list(pix_to_image).count(1) != list(pix_to_image).count(2)

    class TestComputeInversion:

        def test__5_simple_grid__no_sub_grid__sets_up_correct_reconstructor(self):

            pix_grid = np.array([[1.0, 1.0], [-1.0, 1.0], [0.0, 0.0], [1.0, -1.0], [-1.0, -1.0]])
            pix_border = mask.ImageGridBorder(arr=np.array([0, 1, 3, 4]))

            pix_sub_grid = np.array([[1.0, 1.0], [-1.0, 1.0], [0.0, 0.0], [1.0, -1.0], [-1.0, -1.0]])
            pix_sub_border = mask.SubGridBorder(arr=np.array([0, 1, 3, 4]), sub_grid_size=1)

            sub_to_image = np.array([0, 1, 2, 3, 4])

            cluster_to_image = np.array([0, 1, 2, 3, 4])
            image_to_cluster = np.array([0, 1, 2, 3, 4])
            cluster_mask = MockSparseMask(sparse_to_image=cluster_to_image, image_to_sparse=image_to_cluster)

            grids = MockGridCollection(image=pix_grid,
                                       sub=MockSubGridCoords(pix_sub_grid, sub_to_image, sub_grid_size=1))

            borders = MockBorderCollection(image=pix_border, sub=pix_sub_border)

            pix = pixelization.AmorphousRegConst(pixels=5, regularization_coefficients=(1.0,))

            reconstructor = pix.reconstructor_from_pix_grids(grids, borders, cluster_mask)

            assert np.sum(reconstructor.mapping) == 5.0
            assert np.sum(reconstructor.mapping[:, 0]) == 1.0
            assert np.sum(reconstructor.mapping[:, 1]) == 1.0
            assert np.sum(reconstructor.mapping[:, 2]) == 1.0
            assert np.sum(reconstructor.mapping[:, 3]) == 1.0
            assert np.sum(reconstructor.mapping[:, 4]) == 1.0
            assert np.sum(reconstructor.mapping[0, :]) == 1.0
            assert np.sum(reconstructor.mapping[1, :]) == 1.0
            assert np.sum(reconstructor.mapping[2, :]) == 1.0
            assert np.sum(reconstructor.mapping[3, :]) == 1.0
            assert np.sum(reconstructor.mapping[4, :]) == 1.0

            assert np.sum(np.diag(reconstructor.regularization)) == 16.00000005
            assert np.sum(reconstructor.regularization) - np.sum(np.diag(reconstructor.regularization)) == -16.0

            assert set(reconstructor.image_to_pix) == set(np.array([0, 1, 2, 3, 4]))
            assert set(reconstructor.sub_to_pix) == set(np.array([0, 1, 2, 3, 4]))

        def test__15_grid__no_sub_grid__sets_up_correct_reconstructor(self):

            cluster_to_image = np.array([1, 4, 7, 10, 13])
            image_to_cluster = np.array([0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4])
            cluster_mask = MockSparseMask(sparse_to_image=cluster_to_image, image_to_sparse=image_to_cluster)

            pix_grid = np.array([[0.9, 0.9], [1.0, 1.0], [1.1, 1.1],
                                 [-0.9, 0.9], [-1.0, 1.0], [-1.1, 1.1],
                                 [-0.01, 0.01], [0.0, 0.0], [0.01, 0.01],
                                 [0.9, -0.9], [1.0, -1.0], [1.1, -1.1],
                                 [-0.9, -0.9], [-1.0, -1.0], [-1.1, -1.1]])

            pix_border = mask.ImageGridBorder(arr=np.array([2, 5, 11, 14]))

            pix_sub_grid = np.array([[0.9, 0.9], [1.0, 1.0], [1.1, 1.1],
                                     [-0.9, 0.9], [-1.0, 1.0], [-1.1, 1.1],
                                     [-0.01, 0.01], [0.0, 0.0], [0.01, 0.01],
                                     [0.9, -0.9], [1.0, -1.0], [1.1, -1.1],
                                     [-0.9, -0.9], [-1.0, -1.0], [-1.1, -1.1]])

            pix_sub_border = mask.SubGridBorder(arr=np.array([2, 5, 11, 14]))

            sub_to_image = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14])
            grids = MockGridCollection(image=pix_grid,
                                       sub=MockSubGridCoords(pix_sub_grid, sub_to_image,
                                                                   sub_grid_size=1))

            borders = MockBorderCollection(image=pix_border, sub=pix_sub_border)


            pix = pixelization.AmorphousRegConst(pixels=5, regularization_coefficients=(1.0,))

            reconstructor = pix.reconstructor_from_pix_grids(grids, borders, cluster_mask)

            assert np.sum(reconstructor.mapping) == 15.0

            assert np.sum(reconstructor.mapping[:, 0]) == 3.0
            assert np.sum(reconstructor.mapping[:, 1]) == 3.0
            assert np.sum(reconstructor.mapping[:, 2]) == 3.0
            assert np.sum(reconstructor.mapping[:, 3]) == 3.0
            assert np.sum(reconstructor.mapping[:, 4]) == 3.0

            assert np.sum(reconstructor.mapping[0, :]) == 1.0
            assert np.sum(reconstructor.mapping[1, :]) == 1.0
            assert np.sum(reconstructor.mapping[2, :]) == 1.0
            assert np.sum(reconstructor.mapping[3, :]) == 1.0
            assert np.sum(reconstructor.mapping[4, :]) == 1.0
            assert np.sum(reconstructor.mapping[5, :]) == 1.0
            assert np.sum(reconstructor.mapping[6, :]) == 1.0
            assert np.sum(reconstructor.mapping[7, :]) == 1.0
            assert np.sum(reconstructor.mapping[8, :]) == 1.0
            assert np.sum(reconstructor.mapping[9, :]) == 1.0
            assert np.sum(reconstructor.mapping[10, :]) == 1.0
            assert np.sum(reconstructor.mapping[11, :]) == 1.0
            assert np.sum(reconstructor.mapping[12, :]) == 1.0
            assert np.sum(reconstructor.mapping[13, :]) == 1.0
            assert np.sum(reconstructor.mapping[14, :]) == 1.0

            assert np.sum(np.diag(reconstructor.regularization)) == 16.00000005
            assert np.sum(reconstructor.regularization) - np.sum(np.diag(reconstructor.regularization)) == -16.0

            assert set(reconstructor.image_to_pix) == set(np.array([0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4]))
            assert set(reconstructor.sub_to_pix) == set(np.array([0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4]))

        def test__5_simple_grid__include_sub_grid__sets_up_correct_reconstructor(self):

            pix_grid = np.array([[1.0, 1.0], [-1.0, 1.0], [0.0, 0.0], [1.0, -1.0], [-1.0, -1.0]])
            pix_border = mask.ImageGridBorder(arr=np.array([0, 1, 3, 4]))

            cluster_to_image = np.array([0, 1, 2, 3, 4])
            image_to_cluster = np.array([0, 1, 2, 3, 4])
            cluster_mask = MockSparseMask(sparse_to_image=cluster_to_image, image_to_sparse=image_to_cluster)

            pix_sub_grid = np.array([[1.0, 1.0], [1.0, 1.0], [1.0, 1.0], [0.0, 0.0],
                                     [-1.0, 1.0], [-1.0, 1.0], [-1.0, 1.0], [0.0, 0.0],
                                     [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0],
                                     [1.0, -1.0], [1.0, -1.0], [1.0, -1.0], [0.0, 0.0],
                                     [-1.0, -1.0], [-1.0, -1.0], [-1.0, -1.0], [0.0, 0.0]])

            pix_sub_border = mask.SubGridBorder(arr=np.array([0, 1, 2, 4, 5, 6, 12, 13, 14, 16, 17, 18]))

            sub_to_image = np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4])

            grids = MockGridCollection(image=pix_grid,
                                       sub=MockSubGridCoords(pix_sub_grid, sub_to_image,
                                                                   sub_grid_size=2))

            borders = MockBorderCollection(image=pix_border, sub=pix_sub_border)

            pix = pixelization.AmorphousRegConst(pixels=5, regularization_coefficients=(1.0,))

            reconstructor = pix.reconstructor_from_pix_grids(grids, borders, cluster_mask)

            assert np.sum(reconstructor.mapping) == 5.0

            assert np.sum(reconstructor.mapping[0, :]) == 1.0
            assert np.sum(reconstructor.mapping[1, :]) == 1.0
            assert np.sum(reconstructor.mapping[2, :]) == 1.0
            assert np.sum(reconstructor.mapping[3, :]) == 1.0
            assert np.sum(reconstructor.mapping[4, :]) == 1.0

            assert np.sum(reconstructor.mapping[:, 0]) or np.sum(reconstructor.mapping[:, 1]) or np.sum(
                reconstructor.mapping[:, 2]) or np.sum(reconstructor.mapping[:, 3]) or np.sum(
                reconstructor.mapping[:, 4]) == 0.75

            assert np.sum(np.diag(reconstructor.regularization)) == 16.00000005
            assert np.sum(reconstructor.regularization) - np.sum(np.diag(reconstructor.regularization)) == -16.0

            assert set(reconstructor.image_to_pix) == set(np.array([0, 1, 2, 3, 4]))
            assert set(reconstructor.sub_to_pix) == set(
                np.array([0, 0, 0, 2, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 2, 4, 4, 4, 2]))

        def test__same_as_above_but_grid_requires_border_relocation(self):


            pix_grid = np.array([[1.0, 1.0], [-1.0, 1.0], [0.0, 0.0], [1.0, -1.0], [-1.0, -1.0]])
            pix_border = mask.ImageGridBorder(arr=np.array([0, 1, 3, 4]))

            cluster_to_image = np.array([0, 1, 2, 3, 4])
            image_to_cluster = np.array([0, 1, 2, 3, 4])
            cluster_mask = MockSparseMask(sparse_to_image=cluster_to_image, image_to_sparse=image_to_cluster)

            pix_sub_grid = np.array([[1.0, 1.0], [1.0, 1.0], [1.0, 1.0], [0.0, 0.0],
                                     [-1.0, 1.0], [-1.0, 1.0], [-1.0, 1.0], [0.0, 0.0],
                                     [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0],
                                     [1.0, -1.0], [1.0, -1.0], [1.0, -1.0], [0.0, 0.0],
                                     [-1.0, -1.0], [-1.0, -1.0], [-1.0, -1.0], [0.0, 0.0]])

            pix_sub_border = mask.SubGridBorder(arr=np.array([0, 4, 12, 16]))

            sub_to_image = np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4])

            grids = MockGridCollection(image=pix_grid,
                                       sub=MockSubGridCoords(pix_sub_grid, sub_to_image,
                                                                   sub_grid_size=2))

            borders = MockBorderCollection(image=pix_border, sub=pix_sub_border)

            pix = pixelization.AmorphousRegConst(pixels=5, regularization_coefficients=(1.0,))

            reconstructor = pix.reconstructor_from_pix_grids(grids, borders, cluster_mask)

            assert np.sum(reconstructor.mapping) == 5.0

            assert np.sum(reconstructor.mapping[0, :]) == 1.0
            assert np.sum(reconstructor.mapping[1, :]) == 1.0
            assert np.sum(reconstructor.mapping[2, :]) == 1.0
            assert np.sum(reconstructor.mapping[3, :]) == 1.0
            assert np.sum(reconstructor.mapping[4, :]) == 1.0

            assert np.sum(reconstructor.mapping[:, 0]) or np.sum(reconstructor.mapping[:, 1]) or np.sum(
                reconstructor.mapping[:, 2]) or np.sum(reconstructor.mapping[:, 3]) or np.sum(
                reconstructor.mapping[:, 4]) == 0.75

            assert np.sum(np.diag(reconstructor.regularization)) == 16.00000005
            assert np.sum(reconstructor.regularization) - np.sum(np.diag(reconstructor.regularization)) == -16.0

            assert set(reconstructor.image_to_pix) == set(np.array([0, 1, 2, 3, 4]))
            assert set(reconstructor.sub_to_pix) == set(
                np.array([0, 0, 0, 2, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 2, 4, 4, 4, 2]))


            


from profiling import profiling_data
from profiling import tools
from analysis import ray_tracing
from analysis import galaxy
from profiles import mass_profiles
from pixelization import pixelization
from autolens import exc
import numpy as np
import pytest
import numba


sub_grid_size = 4
psf_size = (41, 41)

sie = mass_profiles.EllipticalIsothermal(centre=(0.010, 0.032), einstein_radius=1.47, axis_ratio=0.849, phi=73.6)
shear = mass_profiles.ExternalShear(magnitude=0.0663, phi=160.5)
lens_galaxy = galaxy.Galaxy(mass_profile_0=sie, mass_profile_1=shear)

source_pix = galaxy.Galaxy(pixelization=pixelization.RectangularRegConst(shape=(36, 36)))

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, sub_grid_size=sub_grid_size, psf_shape=psf_size)
euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, sub_grid_size=sub_grid_size, psf_shape=psf_size)
hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, sub_grid_size=sub_grid_size, psf_shape=psf_size)
hst_up = profiling_data.setup_class(name='HSTup', pixel_scale=0.03, sub_grid_size=sub_grid_size, psf_shape=psf_size)
# ao = profiling_data.setup_class(name='AO', pixel_scale=0.01, sub_grid_size=sub_grid_size, psf_shape=psf_size)

lsst_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_pix],
                                 image_plane_grids=lsst.grids)
euclid_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_pix],
                                   image_plane_grids=euclid.grids)
hst_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_pix], image_plane_grids=hst.grids)
hst_up_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_pix],
                                   image_plane_grids=hst_up.grids)
# ao_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_pix], image_plane_grids=ao.grids)

lsst_recon = lsst_tracer.reconstructors_from_source_plane(lsst.borders, cluster_mask=None)
euclid_recon = euclid_tracer.reconstructors_from_source_plane(euclid.borders, cluster_mask=None)
hst_recon = hst_tracer.reconstructors_from_source_plane(hst.borders, cluster_mask=None)
hst_up_recon = hst_up_tracer.reconstructors_from_source_plane(hst_up.borders, cluster_mask=None)
# ao_recon = ao_tracer.reconstructors_from_source_plane(ao.borders, cluster_mask=None)

lsst_reconstructed = lsst_recon.reconstruct_image(lsst.masked_image, lsst.masked_image.background_noise,
                                                  lsst.masked_image.convolver_mapping_matrix)
euclid_reconstructed = euclid_recon.reconstruct_image(euclid.masked_image, euclid.masked_image.background_noise,
                                                  euclid.masked_image.convolver_mapping_matrix)
hst_reconstructed = hst_recon.reconstruct_image(hst.masked_image, hst.masked_image.background_noise,
                                                  hst.masked_image.convolver_mapping_matrix)
hst_up_reconstructed = hst_up_recon.reconstruct_image(hst_up.masked_image, hst_up.masked_image.background_noise,
                                                  hst_up.masked_image.convolver_mapping_matrix)
# ao_reconstructed = ao_recon.reconstruct_image(ao.masked_image, ao.masked_image.noise,
#                                                   ao.masked_image.convolver_mapping_matrix)

@tools.tick_toc_x1
def lsst_solution():
    lsst_reconstructed.regularization_term_from_reconstruction()

@tools.tick_toc_x1
def euclid_solution():
    euclid_reconstructed.regularization_term_from_reconstruction()

@tools.tick_toc_x1
def hst_solution():
    hst_reconstructed.regularization_term_from_reconstruction()

@tools.tick_toc_x1
def hst_up_solution():
    hst_up_reconstructed.regularization_term_from_reconstruction()

@tools.tick_toc_x1
def ao_solution():
    ao_reconstructed.regularization_term_from_reconstruction()

if __name__ == "__main__":
    lsst_solution()
    euclid_solution()
    hst_solution()
    hst_up_solution()
    ao_solution()


from autolens.pipeline import pipeline as pl
from autolens.pipeline import phase as ph
from autolens.imaging import image as im
from autolens.imaging import scaled_array
from autolens.analysis import galaxy_prior as gp
from autolens.profiles import mass_profiles
from autolens.autopipe import model_mapper as mm
from autolens.autopipe import non_linear as nl
import shutil
import numpy as np

import os

dirpath = os.path.dirname(os.path.realpath(__file__))


def load_image(name):
    # Load up the weighted_data
    data_dir = "{}/../../data/{}".format(dirpath, name)

    data = scaled_array.ScaledArray.from_fits(file_path=data_dir + '/image', hdu=0, pixel_scale=0.1)
    noise = scaled_array.ScaledArray.from_fits(file_path=data_dir + '/noise', hdu=0, pixel_scale=0.1)
    psf = im.PSF.from_fits(file_path=data_dir + '/psf', hdu=0, pixel_scale=0.1)

    return im.Image(array=data, pixel_scale=0.05, psf=psf, noise=noise)


def test_source_only_phase_1():
    phase1 = pl.make_source_only_pipeline().phases[0]

    result = phase1.run(load_image('source_sersic'))
    print(result)


def test_source_only_phase_2():
    lens_galaxy = gp.GalaxyPrior(
        sie=mass_profiles.SphericalIsothermal,
        shear=mass_profiles.ExternalShear)

    variable = mm.ModelMapper()
    variable.lens_galaxy = lens_galaxy

    last_result = ph.SourceLensPhase.Result(mm.ModelInstance(), 1., variable,
                                            galaxy_images=[np.ones(1264), np.ones(1264)])

    phase2 = pl.make_source_only_pipeline().phases[1]

    result = phase2.run(mass_profiles, last_result)
    print(result)


def test_source_only_pipeline():
    pipeline = pl.make_source_only_pipeline()
    results = pipeline.run(load_image('source_sersic'))
    for result in results:
        print(result)


def test_profile_pipeline():
    name = "test_pipeline"
    try:
        shutil.rmtree("{}/../../output/{}".format(dirpath, name))
    except FileNotFoundError:
        pass
    pipeline = pl.make_profile_pipeline(name, optimizer_class=nl.DownhillSimplex)
    results = pipeline.run(load_image("integration/hst_0"))
    for result in results:
        print(result)


if __name__ == "__main__":
    test_profile_pipeline()



from __future__ import division, print_function

import pytest
from autolens.profiles import light_profiles
import math
import numpy as np


@pytest.fixture(name='circular')
def circular_sersic():
    return light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6,
                                           sersic_index=4.0)


@pytest.fixture(name='elliptical')
def elliptical_sersic():
    return light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=1.0, effective_radius=0.6,
                                           sersic_index=4.0)


@pytest.fixture(name='vertical')
def vertical_sersic():
    return light_profiles.EllipticalSersic(axis_ratio=0.5, phi=90.0, intensity=1.0, effective_radius=0.6,
                                           sersic_index=4.0)



class TestConstructors(object):

    def test__setup_sersic(self):
        sersic = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6,
                                                 sersic_index=4.0)

        assert sersic.x_cen == 0.0
        assert sersic.y_cen == 0.0
        assert sersic.axis_ratio == 1.0
        assert sersic.phi == 0.0
        assert sersic.intensity == 1.0
        assert sersic.effective_radius == 0.6
        assert sersic.sersic_index == 4.0
        assert sersic.sersic_constant == pytest.approx(7.66925, 1e-3)
        assert sersic.elliptical_effective_radius == 0.6

    def test__setup_exponential(self):
        exponential = light_profiles.EllipticalExponential(axis_ratio=0.5, phi=0.0, intensity=1.0, effective_radius=0.6)

        assert exponential.x_cen == 0.0
        assert exponential.y_cen == 0.0
        assert exponential.axis_ratio == 0.5
        assert exponential.phi == 0.0
        assert exponential.intensity == 1.0
        assert exponential.effective_radius == 0.6
        assert exponential.sersic_index == 1.0
        assert exponential.sersic_constant == pytest.approx(1.678378, 1e-3)
        assert exponential.elliptical_effective_radius == 0.6 / math.sqrt(0.5)

    def test__setup_dev_vaucouleurs(self):
        dev_vaucouleurs = light_profiles.EllipticalDevVaucouleurs(axis_ratio=0.6, phi=10.0, intensity=2.0,
                                                                  effective_radius=0.9,
                                                                  centre=(0.0, 0.1))

        assert dev_vaucouleurs.x_cen == 0.0
        assert dev_vaucouleurs.y_cen == 0.1
        assert dev_vaucouleurs.axis_ratio == 0.6
        assert dev_vaucouleurs.phi == 10.0
        assert dev_vaucouleurs.intensity == 2.0
        assert dev_vaucouleurs.effective_radius == 0.9
        assert dev_vaucouleurs.sersic_index == 4.0
        assert dev_vaucouleurs.sersic_constant == pytest.approx(7.66925, 1e-3)
        assert dev_vaucouleurs.elliptical_effective_radius == 0.9 / math.sqrt(0.6)

    def test__setup_core_sersic(self):
        cored_sersic = light_profiles.EllipticalCoreSersic(axis_ratio=0.5, phi=0.0, intensity=1.0,
                                                           effective_radius=5.0, sersic_index=4.0, radius_break=0.01,
                                                           intensity_break=0.1, gamma=1.0, alpha=1.0)

        assert cored_sersic.x_cen == 0.0
        assert cored_sersic.y_cen == 0.0
        assert cored_sersic.axis_ratio == 0.5
        assert cored_sersic.phi == 0.0
        assert cored_sersic.intensity == 1.0
        assert cored_sersic.effective_radius == 5.0
        assert cored_sersic.sersic_index == 4.0
        assert cored_sersic.sersic_constant == pytest.approx(7.66925, 1e-3)
        assert cored_sersic.radius_break == 0.01
        assert cored_sersic.intensity_break == 0.1
        assert cored_sersic.gamma == 1.0
        assert cored_sersic.alpha == 1.0
        assert cored_sersic.elliptical_effective_radius == 5.0 / math.sqrt(0.5)

    def test_component_numbers_four_profiles(self):
        # TODO : Perform Counting reset better

        from itertools import count

        light_profiles.EllipticalLightProfile._ids = count()

        sersic_0 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6,
                                                   sersic_index=4.0)

        sersic_1 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6,
                                                   sersic_index=4.0)

        sersic_2 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6,
                                                   sersic_index=4.0)

        sersic_3 = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6,
                                                   sersic_index=4.0)

        assert sersic_0.component_number == 0
        assert sersic_1.component_number == 1
        assert sersic_2.component_number == 2
        assert sersic_3.component_number == 3


class TestProfiles(object):
    
    class TestSersic:

        def test__intensity_at_radius__correct_value(self):
            sersic = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6,
                                                     sersic_index=4.0)

            intensity = sersic.intensity_at_radius(radius=1.0)
            assert intensity == pytest.approx(0.351797, 1e-3)

        def test__intensity_at_radius_2__correct_value(self):
            sersic = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=2.0)
            intensity = sersic.intensity_at_radius(
                radius=1.5)  # 3.0 * exp(-3.67206544592 * (1,5/2.0) ** (1.0 / 2.0)) - 1) = 0.351797
            assert intensity == pytest.approx(4.90657319276, 1e-3)

        def test__intensity_from_coordinate_grid__different_axis_ratio(self):
            sersic = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=2.0)

            intensity = sersic.intensity_from_grid(grid=np.array([[0.0, 1.0]]))

            assert intensity == pytest.approx(5.38066670129, 1e-3)

        def test__intensity_from_coordinate_grid__different_rotate_phi_90_same_result(self):
            sersic = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=2.0)

            intensity_1 = sersic.intensity_from_grid(grid=np.array([[0.0, 1.0]]))

            sersic = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=90.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=2.0)

            intensity_2 = sersic.intensity_from_grid(grid=np.array([[1.0, 0.0]]))

            assert intensity_1 == intensity_2

    class TestExponential:

        def test__intensity_at_radius__correct_value(self):
            exponential = light_profiles.EllipticalExponential(axis_ratio=1.0, phi=0.0, intensity=1.0,
                                                               effective_radius=0.6)

            intensity = exponential.intensity_at_radius(radius=1.0)
            assert intensity == pytest.approx(0.3266, 1e-3)

        def test__intensity_at_radius_2__correct_value(self):
            exponential = light_profiles.EllipticalExponential(axis_ratio=1.0, phi=0.0, intensity=3.0,
                                                               effective_radius=2.0)
            intensity = exponential.intensity_at_radius(radius=1.5)
            assert intensity == pytest.approx(4.5640, 1e-3)

        def test__intensity_from_coordinate_grid_1(self):
            exponential = light_profiles.EllipticalExponential(axis_ratio=0.5, phi=0.0, intensity=3.0,
                                                               effective_radius=2.0)

            intensity = exponential.intensity_from_grid(grid=np.array([[0.0, 1.0]]))

            assert intensity == pytest.approx(4.9047, 1e-3)

        def test__intensity_from_coordinate_grid_2(self):
            exponential = light_profiles.EllipticalExponential(axis_ratio=0.5, phi=90.0, intensity=2.0,
                                                               effective_radius=3.0)

            intensity = exponential.intensity_from_grid(grid=np.array([[1.0, 0.0]]))

            assert intensity == pytest.approx(4.8566, 1e-3)

        def test__double_intensity_doubles_intensity_value(self):
            exponential = light_profiles.EllipticalExponential(axis_ratio=0.5, phi=90.0, intensity=4.0,
                                                               effective_radius=3.0)

            intensity = exponential.intensity_from_grid(grid=np.array([[1.0, 0.0]]))

            assert intensity == pytest.approx(2.0 * 4.8566, 1e-3)

        def test__intensity_from_coordinate_grid__different_rotate_phi_90_same_result(self):
            exponential = light_profiles.EllipticalExponential(axis_ratio=0.5, phi=0.0, intensity=3.0,
                                                               effective_radius=2.0)

            intensity_1 = exponential.intensity_from_grid(grid=np.array([[0.0, 1.0]]))

            exponential = light_profiles.EllipticalExponential(axis_ratio=0.5, phi=90.0, intensity=3.0,
                                                               effective_radius=2.0)

            intensity_2 = exponential.intensity_from_grid(grid=np.array([[1.0, 0.0]]))

            assert intensity_1 == intensity_2

    class TestDevVaucouleurs:

        def test__intensity_at_radius__correct_value(self):
            dev_vaucouleurs = light_profiles.EllipticalDevVaucouleurs(axis_ratio=1.0, phi=0.0, intensity=1.0,
                                                                      effective_radius=0.6)

            intensity = dev_vaucouleurs.intensity_at_radius(radius=1.0)
            assert intensity == pytest.approx(0.3518, 1e-3)

        def test__intensity_at_radius_2__correct_value(self):
            dev_vaucouleurs = light_profiles.EllipticalDevVaucouleurs(axis_ratio=1.0, phi=0.0, intensity=3.0,
                                                                      effective_radius=2.0)
            intensity = dev_vaucouleurs.intensity_at_radius(radius=1.5)
            assert intensity == pytest.approx(5.1081, 1e-3)

        def test__intensity_from_coordinate_grid_1(self):
            dev_vaucouleurs = light_profiles.EllipticalDevVaucouleurs(axis_ratio=0.5, phi=0.0, intensity=3.0,
                                                                      effective_radius=2.0)

            intensity = dev_vaucouleurs.intensity_from_grid(grid=np.array([[0.0, 1.0]]))

            assert intensity == pytest.approx(5.6697, 1e-3)

        def test__intensity_from_coordinate_grid_2(self):
            dev_vaucouleurs = light_profiles.EllipticalDevVaucouleurs(axis_ratio=0.5, phi=90.0, intensity=2.0,
                                                                      effective_radius=3.0)

            intensity = dev_vaucouleurs.intensity_from_grid(grid=np.array([[1.0, 0.0]]))

            assert intensity == pytest.approx(7.4455, 1e-3)

        def test__double_intensity_doubles_intensity_value(self):
            dev_vaucouleurs = light_profiles.EllipticalDevVaucouleurs(axis_ratio=0.5, phi=90.0, intensity=4.0,
                                                                      effective_radius=3.0)

            intensity = dev_vaucouleurs.intensity_from_grid(grid=np.array([[1.0, 0.0]]))

            assert intensity == pytest.approx(2.0 * 7.4455, 1e-3)

        def test__intensity_from_coordinate_grid__different_rotate_phi_90_same_result(self):
            dev_vaucouleurs = light_profiles.EllipticalDevVaucouleurs(axis_ratio=0.5, phi=0.0, intensity=3.0,
                                                                      effective_radius=2.0)

            intensity_1 = dev_vaucouleurs.intensity_from_grid(grid=np.array([[0.0, 1.0]]))

            dev_vaucouleurs = light_profiles.EllipticalDevVaucouleurs(axis_ratio=0.5, phi=90.0, intensity=3.0,
                                                                      effective_radius=2.0)

            intensity_2 = dev_vaucouleurs.intensity_from_grid(grid=np.array([[1.0, 0.0]]))

            assert intensity_1 == intensity_2

    class TestCoreSersic(object):

        def test__intensity_at_radius__correct_value(self):
            core_sersic = light_profiles.EllipticalCoreSersic(axis_ratio=0.5, phi=0.0, intensity=1.0,
                                                              effective_radius=5.0, sersic_index=4.0,
                                                              radius_break=0.01,
                                                              intensity_break=0.1, gamma=1.0, alpha=1.0)

            assert core_sersic.intensity_at_radius(0.01) == 0.1


class TestLuminosityIntegral(object):
    class TestWithinCircle(object):

        def test__spherical_exponential__compare_to_analytic(self):

            import math
            import scipy.special

            sersic = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=1.0)

            integral_radius = 5.5

            # Use gamma function for analytic computation of the intensity within a radius=0.5

            x = sersic.sersic_constant * (integral_radius / sersic.effective_radius) ** (1.0 / sersic.sersic_index)

            intensity_analytic = sersic.intensity * sersic.effective_radius ** 2 * 2 * math.pi * sersic.sersic_index * (
                    math.e ** sersic.sersic_constant / (
                    sersic.sersic_constant ** (2 * sersic.sersic_index))) * scipy.special.gamma(
                2 * sersic.sersic_index) * scipy.special.gammainc(
                2 * sersic.sersic_index, x)

            intensity_integral = sersic.luminosity_within_circle(radius=integral_radius)

            assert intensity_analytic == pytest.approx(intensity_integral, 1e-3)

        def test__spherical_sersic_index_2__compare_to_analytic(self):

            import math
            import scipy.special

            sersic = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=2.0)

            integral_radius = 0.5

            # Use gamma function for analytic computation of the intensity within a radius=0.5

            x = sersic.sersic_constant * ((integral_radius / sersic.effective_radius) ** (1.0 / sersic.sersic_index))

            intensity_analytic = sersic.intensity * sersic.effective_radius ** 2 * 2 * math.pi * sersic.sersic_index * \
                                 ((math.e ** sersic.sersic_constant) / (
                                         sersic.sersic_constant ** (2 * sersic.sersic_index))) * \
                                 scipy.special.gamma(2 * sersic.sersic_index) * scipy.special.gammainc(
                2 * sersic.sersic_index, x)

            intensity_integral = sersic.luminosity_within_circle(radius=0.5)

            assert intensity_analytic == pytest.approx(intensity_integral, 1e-3)

        def test__spherical_dev_vaucouleurs__compare_to_analytic(self):

            import math
            import scipy.special

            sersic = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=4.0)

            integral_radius = 0.5

            # Use gamma function for analytic computation of the intensity within a radius=0.5

            x = sersic.sersic_constant * ((integral_radius / sersic.effective_radius) ** (1.0 / sersic.sersic_index))

            intensity_analytic = sersic.intensity * sersic.effective_radius ** 2 * 2 * math.pi * sersic.sersic_index * \
                                 ((math.e ** sersic.sersic_constant) / (
                                         sersic.sersic_constant ** (2 * sersic.sersic_index))) * \
                                 scipy.special.gamma(2 * sersic.sersic_index) * scipy.special.gammainc(
                2 * sersic.sersic_index, x)

            intensity_integral = sersic.luminosity_within_circle(radius=0.5)

            assert intensity_analytic == pytest.approx(intensity_integral, 1e-3)

        def test__spherical_exponential__compare_to_grid(self):

            sersic = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=1.0)

            import numpy as np

            integral_radius = 1.0
            luminosity_tot = 0.0

            xs = np.linspace(-1.5, 1.5, 40)
            ys = np.linspace(-1.5, 1.5, 40)

            edge = xs[1] - xs[0]
            area = edge ** 2

            for x in xs:
                for y in ys:

                    eta = math.sqrt(x ** 2 + y ** 2)

                    if eta < integral_radius:
                        luminosity_tot += sersic.intensity_at_radius(eta) * area

            intensity_integral = sersic.luminosity_within_circle(radius=integral_radius)

            assert luminosity_tot == pytest.approx(intensity_integral, 0.02)

        def test__spherical_sersic_2__compare_to_grid(self):

            sersic = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=2.0)

            import numpy as np

            integral_radius = 1.0
            luminosity_tot = 0.0

            xs = np.linspace(-1.5, 1.5, 40)
            ys = np.linspace(-1.5, 1.5, 40)

            edge = xs[1] - xs[0]
            area = edge ** 2

            for x in xs:
                for y in ys:

                    eta = math.sqrt(x ** 2 + y ** 2)
                    if eta < integral_radius:
                        luminosity_tot += sersic.intensity_at_radius(eta) * area

            intensity_integral = sersic.luminosity_within_circle(radius=integral_radius)

            assert luminosity_tot == pytest.approx(intensity_integral, 0.02)

        def test__spherical_dev_vaucauleurs__compare_to_grid(self):

            sersic = light_profiles.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=4.0)

            import numpy as np

            integral_radius = 1.0
            luminosity_tot = 0.0

            xs = np.linspace(-1.5, 1.5, 40)
            ys = np.linspace(-1.5, 1.5, 40)

            edge = xs[1] - xs[0]
            area = edge ** 2

            for x in xs:
                for y in ys:

                    eta = math.sqrt(x ** 2 + y ** 2)
                    if eta < integral_radius:
                        luminosity_tot += sersic.intensity_at_radius(eta) * area

            intensity_integral = sersic.luminosity_within_circle(radius=integral_radius)

            assert luminosity_tot == pytest.approx(intensity_integral, 0.02)

        def test__elliptical_exponential__compare_to_grid(self):

            sersic = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=1.0)

            integral_radius = 0.5
            luminosity_tot = 0.0

            xs = np.linspace(-1.0, 1.0, 40)
            ys = np.linspace(-1.0, 1.0, 40)

            edge = xs[1] - xs[0]
            area = edge ** 2

            for x in xs:
                for y in ys:

                    eta = math.sqrt(x ** 2 + y ** 2)
                    if eta < integral_radius:
                        luminosity_tot += sersic.intensity_at_radius(eta) * area

            intensity_integral = sersic.luminosity_within_circle(radius=integral_radius)

            assert luminosity_tot == pytest.approx(intensity_integral, 0.02)

        def test__elliptical_sersic_2__compare_to_grid(self):

            sersic = light_profiles.EllipticalSersic(axis_ratio=0.3, phi=0.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=2.0)

            integral_radius = 0.5
            luminosity_tot = 0.0

            xs = np.linspace(-1.8, 1.8, 50)
            ys = np.linspace(-1.8, 1.8, 50)

            edge = xs[1] - xs[0]
            area = edge ** 2

            for x in xs:
                for y in ys:

                    eta = math.sqrt(x ** 2 + y ** 2)
                    if eta < integral_radius:
                        luminosity_tot += sersic.intensity_at_radius(eta) * area

            intensity_integral = sersic.luminosity_within_circle(radius=integral_radius)

            assert luminosity_tot == pytest.approx(intensity_integral, 0.02)

    class TestWithinEllipse(object):

        def test__elliptical_exponential__compare_to_grid(self):

            sersic = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=1.0)

            integral_radius = 0.5
            luminosity_tot = 0.0

            xs = np.linspace(-1.0, 1.0, 40)
            ys = np.linspace(-1.0, 1.0, 40)

            edge = xs[1] - xs[0]
            area = edge ** 2

            for x in xs:
                for y in ys:

                    eta = sersic.grid_to_elliptical_radius(np.array([[x, y]]))

                    if eta < integral_radius:
                        luminosity_tot += sersic.intensity_at_radius(eta) * area

            intensity_integral = sersic.luminosity_within_ellipse(major_axis=integral_radius)

            assert luminosity_tot == pytest.approx(intensity_integral, 0.02)

        def test__elliptical_sersic_2__compare_to_grid(self):

            sersic = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=90.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=2.0)

            integral_radius = 0.5
            luminosity_tot = 0.0

            xs = np.linspace(-1.8, 1.8, 80)
            ys = np.linspace(-1.8, 1.8, 80)

            edge = xs[1] - xs[0]
            area = edge ** 2

            for x in xs:
                for y in ys:

                    eta = sersic.grid_to_elliptical_radius(np.array([[x, y]]))

                    if eta < integral_radius:
                        luminosity_tot += sersic.intensity_at_radius(eta) * area

            intensity_integral = sersic.luminosity_within_ellipse(major_axis=integral_radius)

            assert luminosity_tot == pytest.approx(intensity_integral, 0.02)

        def test__elliptical_dev_vaucauleurs__compare_to_grid(self):

            sersic = light_profiles.EllipticalSersic(axis_ratio=0.7, phi=30.0, intensity=3.0, effective_radius=2.0,
                                                     sersic_index=4.0)

            integral_radius = 0.5
            luminosity_tot = 0.0

            xs = np.linspace(-1.5, 1.5, 50)
            ys = np.linspace(-1.5, 1.5, 50)

            edge = xs[1] - xs[0]
            area = edge ** 2

            for x in xs:
                for y in ys:

                    eta = sersic.grid_to_elliptical_radius(np.array([[x, y]]))

                    if eta < integral_radius:
                        luminosity_tot += sersic.intensity_at_radius(eta) * area

            intensity_integral = sersic.luminosity_within_ellipse(major_axis=integral_radius)

            assert luminosity_tot == pytest.approx(intensity_integral, 0.01)


class TestCoordinates(object):
    def test__grid_to_eccentric_radius(self, elliptical):
        assert elliptical.grid_to_eccentric_radii(np.array([[1, 1]])) == pytest.approx(
            elliptical.grid_to_eccentric_radii(np.array([[-1, -1]])), 1e-10)

    def test__intensity_from_coordinate_grid(self, elliptical):
        assert elliptical.intensity_from_grid(np.array([[1, 1]])) == \
               pytest.approx(elliptical.intensity_from_grid(np.array([[-1, -1]])), 1e-4)




class HyperImage(object):

    def __init__(self, background_sky_scale=0.0, background_noise_scale=0.0):
        """Class for scaling the noise of different components in an image, primarily the different galaxies (e.g. the \
        lens, source).

        Parameters
        -----------
        background_sky_scale : float
            The value by which the background scale is scaled (electrons per second).
        background_noise_scale : float
            The value by which the background noise is increased (electrons per second).
        """
        self.background_sky_scale = background_sky_scale
        self.background_noise_scale = background_noise_scale

    @property
    def subscript(self):
        return 'hi'

    @property
    def parameter_labels(self):
        return [r'\sigma', r'\sigma']

    def sky_scaled_image_from_image(self, image):
        """Compute a new image with the background sky level scaled. This can simply multiple by a constant factor \
        (assuming a uniform background sky) because the image is in units electrons per second.

        Parameters
        -----------
        image : ndarray
            The image before scaling (electrons per second).
        """
        return image + self.background_sky_scale

    def scaled_noise_from_background_noise(self, noise, background_noise):
        """Compute a scaled noise map from the baseline noise map. This scales each galaxy component individually \
        using their galaxy contribution map and sums their scaled noise maps with the baseline and background noise maps.

        Parameters
        -----------
        noise : ndarray
            The noise before scaling (electrons per second)..
        background_noise : ndarray
            The background noise values (electrons per second)..
        """
        return noise + (self.background_noise_scale * background_noise)



from autolens.imaging import mask
from autolens.imaging import image
from autolens.imaging import convolution

import numpy as np
import pytest


@pytest.fixture(name='sim_image_31x31', scope='function')
def sim_grid_9x9():
    sim_grid_9x9.ma = mask.Mask.for_simulate(shape_arc_seconds=(5.5, 5.5), pixel_scale=0.5, psf_size=(3, 3))
    sim_grid_9x9.image_grid = sim_grid_9x9.ma.coordinates_collection_for_subgrid_size_and_blurring_shape(
        sub_grid_size=1,
        blurring_shape=(3, 3))
    sim_grid_9x9.mapping = sim_grid_9x9.ma.grid_mapping_with_sub_grid_size(sub_grid_size=1, cluster_grid_size=1)
    return sim_grid_9x9


class TestConvolutuion:

    def test__compare_convolver_to_2d_convolution(self):
        # Setup a blurred image, using the PSF to perform the convolution in 2D, then mask it to make a 1d array.

        im = np.arange(900).reshape(30, 30)
        psf = image.PSF(array=np.arange(49).reshape(7, 7))
        blurred_im = psf.convolve(im)
        msk = mask.Mask.circular(shape_arc_seconds=(30.0, 30.0), pixel_scale=1.0, radius_mask=4.0)
        blurred_masked_im_0 = msk.map_to_1d(blurred_im)

        # Now reproduce this image using the frame convolver_image

        blurring_mask = msk.blurring_mask_for_kernel_shape(psf.shape)
        convolver = convolution.ConvolverImage(mask=msk, blurring_mask=blurring_mask, psf=psf)
        im_1d = msk.map_to_1d(im)
        blurring_im_1d = blurring_mask.map_to_1d(im)
        blurred_masked_im_1 = convolver.convolve_image_jit(image_array=im_1d, blurring_array=blurring_im_1d)

        assert blurred_masked_im_0 == pytest.approx(blurred_masked_im_1, 1e-4)



from profiling import profiling_data
from profiling import tools
from profiles import geometry_profiles
import pytest
import numpy as np
import numba

class EllipticalProfile(geometry_profiles.EllipticalProfile):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0):
        """ Generic elliptical profiles class to contain functions shared by light and mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The coordinates of the centre of the profiles
        axis_ratio : float
            Ratio of profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        """
        super(EllipticalProfile, self).__init__(centre)
        self.axis_ratio = axis_ratio
        self.phi = phi

    def grid_angle_to_profile(self, theta_grid):
        theta_coordinate_to_profile = np.add(theta_grid, - self.phi_radians)
        return np.cos(theta_coordinate_to_profile), np.sin(theta_coordinate_to_profile)

    def grid_radius_to_cartesian(self, grid, radius):
        theta_grid = np.arctan2(grid[:, 1], grid[:, 0])
        cos_theta, sin_theta = self.grid_angle_to_profile(theta_grid)
        return np.multiply(radius[:, None], np.vstack((cos_theta, sin_theta)).T)

    def grid_radius_to_cartesian_jitted(self, grid, radius):
        return self.grid_radius_to_cartesian_jit(grid, radius, self.phi_radians)

    @staticmethod
    @numba.jit(nopython=True)
    def grid_radius_to_cartesian_jit(grid, radius, phi_radians):

        cartesian = np.zeros(grid.shape)

        for i in range(grid.shape[0]):

            theta_coordinate_to_profile = np.arctan2(grid[i,1], grid[i,0]) - phi_radians
            cartesian[i, 0] = radius[i] * np.cos(theta_coordinate_to_profile)
            cartesian[i, 1] = radius[i] * np.sin(theta_coordinate_to_profile)

        return cartesian

geometry = EllipticalProfile(centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0)

sub_grid_size=4

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, sub_grid_size=sub_grid_size)
euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, sub_grid_size=sub_grid_size)
hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, sub_grid_size=sub_grid_size)
hst_up = profiling_data.setup_class(name='HSTup', pixel_scale=0.03, sub_grid_size=sub_grid_size)
ao = profiling_data.setup_class(name='AO', pixel_scale=0.01, sub_grid_size=sub_grid_size)

lsst_radius = np.ones(lsst.coords.sub_grid_coords.shape[0])
euclid_radius = np.ones(euclid.coords.sub_grid_coords.shape[0])
hst_radius = np.ones(hst.coords.sub_grid_coords.shape[0])
hst_up_radius = np.ones(hst_up.coords.sub_grid_coords.shape[0])
ao_radius = np.ones(ao.coords.sub_grid_coords.shape[0])

assert (geometry.grid_radius_to_cartesian(grid=lsst.coords.sub_grid_coords, radius=lsst_radius) ==
        pytest.approx(geometry.grid_radius_to_cartesian_jitted(grid=lsst.coords.sub_grid_coords, radius=lsst_radius), 1e-4))

@tools.tick_toc_x20
def lsst_solution():
    geometry.grid_radius_to_cartesian_jitted(grid=lsst.coords.sub_grid_coords, radius=lsst_radius)

@tools.tick_toc_x20
def euclid_solution():
    geometry.grid_radius_to_cartesian_jitted(grid=euclid.coords.sub_grid_coords, radius=euclid_radius)

@tools.tick_toc_x20
def hst_solution():
    geometry.grid_radius_to_cartesian_jitted(grid=hst.coords.sub_grid_coords, radius=hst_radius)

@tools.tick_toc_x20
def hst_up_solution():
    geometry.grid_radius_to_cartesian_jitted(grid=hst_up.coords.sub_grid_coords, radius=hst_up_radius)

@tools.tick_toc_x20
def ao_solution():
    geometry.grid_radius_to_cartesian_jitted(grid=ao.coords.sub_grid_coords, radius=ao_radius)

if __name__ == "__main__":
    lsst_solution()
    euclid_solution()
    hst_solution()
    hst_up_solution()
    ao_solution()